# -*- coding: utf-8 -*-
import cherrypy
import datetime
import os
import six

from .model_base import Model, AccessControlledModel
from girder import auditLogger, events
from girder.constants import AccessType, CoreEventHandler
from girder.exceptions import FilePathException, ValidationException
from girder.models.setting import Setting
from girder.settings import SettingKey
from girder.utility import acl_mixin, path as path_util
from girder.utility.model_importer import ModelImporter


class File(acl_mixin.AccessControlMixin, Model):
    """
    This model represents a File, which is stored in an assetstore.
    """

    def initialize(self):
        from girder.utility import assetstore_utilities

        self.name = 'file'
        self.ensureIndices(
            ['itemId', 'assetstoreId', 'exts']
            + assetstore_utilities.fileIndexFields())
        self.ensureTextIndex({'name': 1})
        self.resourceColl = 'item'
        self.resourceParent = 'itemId'

        self.exposeFields(level=AccessType.READ, fields=(
            '_id', 'mimeType', 'itemId', 'exts', 'name', 'created', 'creatorId',
            'size', 'updated', 'linkUrl'))

        self.exposeFields(level=AccessType.SITE_ADMIN, fields=('assetstoreId',))

        events.bind('model.file.save.created',
                    CoreEventHandler.FILE_PROPAGATE_SIZE,
                    self._propagateSizeToItem)

    def remove(self, file, updateItemSize=True, **kwargs):
        """
        Use the appropriate assetstore adapter for whatever assetstore the
        file is stored in, and call deleteFile on it, then delete the file
        record from the database.

        :param file: The file document to remove.
        :param updateItemSize: Whether to update the item size. Only set this
            to False if you plan to delete the item and do not care about
            updating its size.
        """
        from .item import Item

        if file.get('assetstoreId'):
            self.getAssetstoreAdapter(file).deleteFile(file)

        if file['itemId']:
            item = Item().load(file['itemId'], force=True)
            # files that are linkUrls might not have a size field
            if 'size' in file:
                self.propagateSizeChange(item, -file['size'], updateItemSize)

        Model.remove(self, file)

    def download(self, file, offset=0, headers=True, endByte=None,
                 contentDisposition=None, extraParameters=None):
        """
        Use the appropriate assetstore adapter for whatever assetstore the
        file is stored in, and call downloadFile on it. If the file is a link
        file rather than a file in an assetstore, we redirect to it.

        :param file: The file to download.
        :param offset: The start byte within the file.
        :type offset: int
        :param headers: Whether to set headers (i.e. is this an HTTP request
            for a single file, or something else).
        :type headers: bool
        :param endByte: Final byte to download. If ``None``, downloads to the
            end of the file.
        :type endByte: int or None
        :param contentDisposition: Content-Disposition response header
            disposition-type value.
        :type contentDisposition: str or None
        :type extraParameters: str or None
        """
        events.trigger('model.file.download.request', info={
            'file': file,
            'startByte': offset,
            'endByte': endByte})

        auditLogger.info('file.download', extra={
            'details': {
                'fileId': file['_id'],
                'startByte': offset,
                'endByte': endByte,
                'extraParameters': extraParameters
            }
        })

        if file.get('assetstoreId'):
            try:
                fileDownload = self.getAssetstoreAdapter(file).downloadFile(
                    file, offset=offset, headers=headers, endByte=endByte,
                    contentDisposition=contentDisposition,
                    extraParameters=extraParameters)

                def downloadGenerator():
                    for data in fileDownload():
                        yield data
                    if endByte is None or endByte >= file['size']:
                        events.trigger('model.file.download.complete', info={
                            'file': file,
                            'startByte': offset,
                            'endByte': endByte,
                            'redirect': False})
                return downloadGenerator
            except cherrypy.HTTPRedirect:
                events.trigger('model.file.download.complete', info={
                    'file': file,
                    'startByte': offset,
                    'endByte': endByte,
                    'redirect': True})
                raise
        elif file.get('linkUrl'):
            if headers:
                events.trigger('model.file.download.complete', info={
                    'file': file,
                    'startByte': offset,
                    'endByte': endByte,
                    'redirect': True})
                raise cherrypy.HTTPRedirect(file['linkUrl'])
            else:
                endByte = endByte or len(file['linkUrl'])

                def stream():
                    yield file['linkUrl'][offset:endByte]
                    if endByte >= len(file['linkUrl']):
                        events.trigger('model.file.download.complete', info={
                            'file': file,
                            'startByte': offset,
                            'endByte': endByte,
                            'redirect': False})
                return stream
        else:
            raise Exception('File has no known download mechanism.')

    def validate(self, doc):
        if doc.get('assetstoreId') is None:
            if 'linkUrl' not in doc:
                raise ValidationException(
                    'File must have either an assetstore ID or a link URL.',
                    'linkUrl')
            doc['linkUrl'] = doc['linkUrl'].strip()

            if not doc['linkUrl'].startswith(('http:', 'https:')):
                raise ValidationException(
                    'Linked file URL must start with http: or https:.',
                    'linkUrl')
        if doc.get('assetstoreType'):
            # If assetstore model is overridden, make sure it's a valid model
            self._getAssetstoreModel(doc)
        if 'name' not in doc or not doc['name']:
            raise ValidationException('File name must not be empty.', 'name')

        doc['exts'] = [ext.lower() for ext in doc['name'].split('.')[1:]]

        return doc

    def _getAssetstoreModel(self, file):
        from .assetstore import Assetstore

        if file.get('assetstoreType'):
            try:
                if isinstance(file['assetstoreType'], six.string_types):
                    return ModelImporter.model(file['assetstoreType'])
                else:
                    return ModelImporter.model(*file['assetstoreType'])
            except Exception:
                raise ValidationException(
                    'Invalid assetstore type: %s.' % (file['assetstoreType'],))
        else:
            return Assetstore()

    def createLinkFile(self, name, parent, parentType, url, creator, size=None,
                       mimeType=None, reuseExisting=False):
        """
        Create a file that is a link to a URL, rather than something we maintain
        in an assetstore.

        :param name: The local name for the file.
        :type name: str
        :param parent: The parent object for this file.
        :type parent: girder.models.folder or girder.models.item
        :param parentType: The parent type (folder or item)
        :type parentType: str
        :param url: The URL that this file points to
        :param creator: The user creating the file.
        :type creator: dict
        :param size: The size of the file in bytes. (optional)
        :type size: int
        :param mimeType: The mimeType of the file. (optional)
        :type mimeType: str
        :param reuseExisting: If a file with the same name already exists in
            this location, return it rather than creating a new file.
        :type reuseExisting: bool
        """
        from .item import Item

        if parentType == 'folder':
            # Create a new item with the name of the file.
            item = Item().createItem(
                name=name, creator=creator, folder=parent, reuseExisting=reuseExisting)
        elif parentType == 'item':
            item = parent

        existing = None
        if reuseExisting:
            existing = self.findOne({
                'itemId': item['_id'],
                'name': name
            })

        if existing:
            file = existing
        else:
            file = {
                'created': datetime.datetime.utcnow(),
                'itemId': item['_id'],
                'assetstoreId': None,
                'name': name
            }

        file.update({
            'creatorId': creator['_id'],
            'mimeType': mimeType,
            'linkUrl': url
        })
        if size is not None:
            file['size'] = int(size)

        try:
            if existing:
                file = self.updateFile(file)
            else:
                file = self.save(file)
            return file
        except ValidationException:
            if parentType == 'folder':
                Item().remove(item)
            raise

    def propagateSizeChange(self, item, sizeIncrement, updateItemSize=True):
        """
        Propagates a file size change (or file creation) to the necessary
        parents in the hierarchy. Internally, this records subtree size in
        the item, the parent folder, and the root node under which the item
        lives. Should be called anytime a new file is added, a file is
        deleted, or a file size changes.

        :param item: The parent item of the file.
        :type item: dict
        :param sizeIncrement: The change in size to propagate.
        :type sizeIncrement: int
        :param updateItemSize: Whether the item size should be updated. Set to
            False if you plan to delete the item immediately and don't care to
            update its size.
        """
        from .folder import Folder
        from .item import Item

        if updateItemSize:
            # Propagate size up to item
            Item().increment(query={
                '_id': item['_id']
            }, field='size', amount=sizeIncrement, multi=False)

        # Propagate size to direct parent folder
        Folder().increment(query={
            '_id': item['folderId']
        }, field='size', amount=sizeIncrement, multi=False)

        # Propagate size up to root data node
        ModelImporter.model(item['baseParentType']).increment(query={
            '_id': item['baseParentId']
        }, field='size', amount=sizeIncrement, multi=False)

    def createFile(self, creator, item, name, size, assetstore, mimeType=None,
                   saveFile=True, reuseExisting=False, assetstoreType=None):
        """
        Create a new file record in the database.

        :param item: The parent item.
        :param creator: The user creating the file.
        :param assetstore: The assetstore this file is stored in.
        :param name: The filename.
        :type name: str
        :param size: The size of the file in bytes.
        :type size: int
        :param mimeType: The mimeType of the file.
        :type mimeType: str
        :param saveFile: if False, don't save the file, just return it.
        :type saveFile: bool
        :param reuseExisting: If a file with the same name already exists in
            this location, return it rather than creating a new file.
        :type reuseExisting: bool
        :param assetstoreType: If a model other than assetstore will be used to
            initialize the assetstore adapter for this file, use this parameter to
            specify it. If it's a core model, pass its string name. If it's a plugin
            model, use a 2-tuple of the form (modelName, pluginName).
        :type assetstoreType: str or tuple
        """
        if reuseExisting:
            existing = self.findOne({
                'itemId': item['_id'],
                'name': name
            })
            if existing:
                return existing

        file = {
            'created': datetime.datetime.utcnow(),
            'creatorId': creator['_id'],
            'assetstoreId': assetstore['_id'],
            'name': name,
            'mimeType': mimeType,
            'size': size,
            'itemId': item['_id'] if item else None
        }

        if assetstoreType:
            file['assetstoreType'] = assetstoreType

        if saveFile:
            return self.save(file)
        return file

    def _propagateSizeToItem(self, event):
        """
        This callback updates an item's size to include that of a newly-created
        file.

        This generally should not be called or overridden directly. This should
        not be unregistered, as that would cause item, folder, and collection
        sizes to be inaccurate.
        """
        # This task is not performed in "createFile", in case
        # "saveFile==False". The item size should be updated only when it's
        # certain that the file will actually be saved. It is also possible for
        # "model.file.save" to set "defaultPrevented", which would prevent the
        # item from being saved initially.
        from .item import Item

        fileDoc = event.info
        itemId = fileDoc.get('itemId')
        if itemId and fileDoc.get('size'):
            item = Item().load(itemId, force=True)
            self.propagateSizeChange(item, fileDoc['size'])

    def updateFile(self, file):
        """
        Call this when changing properties of an existing file, such as name
        or MIME type. This causes the updated stamp to change, and also alerts
        the underlying assetstore adapter that file information has changed.
        """
        file['updated'] = datetime.datetime.utcnow()
        file = self.save(file)

        if file.get('assetstoreId'):
            self.getAssetstoreAdapter(file).fileUpdated(file)

        return file

    def getAssetstoreAdapter(self, file):
        """
        Return the assetstore adapter for the given file.  Return None if the
        file has no assetstore.
        """
        from girder.utility import assetstore_utilities

        if not file.get('assetstoreId'):
            return None
        assetstore = self._getAssetstoreModel(file).load(file['assetstoreId'])
        return assetstore_utilities.getAssetstoreAdapter(assetstore)

    def copyFile(self, srcFile, creator, item=None):
        """
        Copy a file so that we don't need to duplicate stored data.

        :param srcFile: The file to copy.
        :type srcFile: dict
        :param creator: The user copying the file.
        :param item: a new item to assign this file to (optional)
        :returns: a dict with the new file.
        """
        # Copy the source file's dictionary.  The individual assetstore
        # implementations will need to fix references if they cannot be
        # directly duplicated.
        file = srcFile.copy()
        # Immediately delete the original id so that we get a new one.
        del file['_id']
        file['copied'] = datetime.datetime.utcnow()
        file['copierId'] = creator['_id']
        if item:
            file['itemId'] = item['_id']
        if file.get('assetstoreId'):
            self.getAssetstoreAdapter(file).copyFile(srcFile, file)
        elif file.get('linkUrl'):
            file['linkUrl'] = srcFile['linkUrl']

        return self.save(file)

    def isOrphan(self, file):
        """
        Returns True if this file is orphaned (its item or attached entity is
        missing).

        :param file: The file to check.
        :type file: dict
        """
        if file.get('attachedToId'):
            attachedToType = file.get('attachedToType')
            if isinstance(attachedToType, six.string_types):
                modelType = ModelImporter.model(attachedToType)
            elif isinstance(attachedToType, list) and len(attachedToType) == 2:
                modelType = ModelImporter.model(*attachedToType)
            else:
                # Invalid 'attachedToType'
                return True
            if isinstance(modelType, (acl_mixin.AccessControlMixin,
                                      AccessControlledModel)):
                attachedDoc = modelType.load(
                    file.get('attachedToId'), force=True)
            else:
                attachedDoc = modelType.load(
                    file.get('attachedToId'))
        else:
            from .item import Item
            attachedDoc = Item().load(file.get('itemId'), force=True)
        return not attachedDoc

    def updateSize(self, file):
        """
        Returns the size of this file. Does not currently check the underlying
        assetstore to verify the size.

        :param file: The file.
        :type file: dict
        """
        # TODO: check underlying assetstore for size?
        return file.get('size', 0), 0

    def open(self, file):
        """
        Use this to expose a Girder file as a python file-like object. At the
        moment, this is a read-only interface, the equivalent of opening a
        system file with ``'rb'`` mode. This can also be used as a context
        manager, e.g.:

        >>> with File().open(file) as fh:
        >>>    while True:
        >>>        chunk = fh.read(CHUNK_LEN)
        >>>        if not chunk:
        >>>            break

        Using it this way will automatically close the file handle for you when
        the ``with`` block is left.

        :param file: A Girder file document.
        :type file: dict
        :return: A file-like object containing the bytes of the file.
        :rtype: girder.utility.abstract_assetstore_adapter.FileHandle
        """
        return self.getAssetstoreAdapter(file).open(file)

    def getGirderMountFilePath(self, file, validate=True):
        """
        If possible, get the path of the file on a local girder mount.

        :param file: The file document.
        :param validate: if True, check if the path exists and raise an
            exception if it does not.
        :returns: a girder mount path to the file or None if no such path is
            available.
        """
        mount = Setting().get(SettingKey.GIRDER_MOUNT_INFORMATION)
        if mount:
            path = mount['path'].rstrip('/') + path_util.getResourcePath('file', file, force=True)
            if not validate or os.path.exists(path):
                return path
        if validate:
            raise FilePathException("This file isn't accessible from a Girder mount.")

    def getLocalFilePath(self, file):
        """
        If an assetstore adapter supports it, return a path to the file on the
        local file system.

        :param file: The file document.
        :returns: a local path to the file or None if no such path is known.
        """
        adapter = self.getAssetstoreAdapter(file)
        try:
            return adapter.getLocalFilePath(file)
        except FilePathException as exc:
            try:
                return self.getGirderMountFilePath(file, True)
            except Exception:
                # If getting a Girder mount path throws, raise the original
                # exception
                pass
            raise exc



# -*- coding: utf-8 -*-
from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.api.rest import boundHandler
from girder.models.setting import Setting

from .settings import PluginSettings


@access.user
@boundHandler
@autoDescribeRoute(
    Description('Get list of item licenses.')
    .param('default', 'Whether to return the default list of item licenses.',
           required=False, dataType='boolean', default=False)
)
def getLicenses(self, default):
    if default:
        licenses = Setting().getDefault(PluginSettings.LICENSES)
    else:
        licenses = Setting().get(PluginSettings.LICENSES)

    return licenses



# -*- coding: utf-8 -*-
import bson
from hashlib import sha512
import pymongo
import six
from six import BytesIO
import time
import uuid

from girder import logger
from girder.api.rest import setResponseHeader
from girder.external.mongodb_proxy import MongoProxy
from girder.models import getDbConnection
from girder.exceptions import ValidationException
from girder.models.file import File
from . import _hash_state
from .abstract_assetstore_adapter import AbstractAssetstoreAdapter


# 2MB chunks. Clients must not send any chunks that are smaller than this
# unless they are sending the final chunk.
CHUNK_SIZE = 2097152

# Cache recent connections so we can skip some start up actions
RECENT_CONNECTION_CACHE_TIME = 600  # seconds
RECENT_CONNECTION_CACHE_MAX_SIZE = 100
_recentConnections = {}


def _ensureChunkIndices(collection):
    """
    Ensure that we have appropriate indices on the chunk collection.

    :param collection: a connection to a mongo collection.
    """
    collection.create_index([
        ('uuid', pymongo.ASCENDING),
        ('n', pymongo.ASCENDING)
    ], unique=True)


class GridFsAssetstoreAdapter(AbstractAssetstoreAdapter):
    """
    This assetstore type stores files within MongoDB using the GridFS data
    model.
    """

    @staticmethod
    def validateInfo(doc):
        """
        Validate the assetstore -- make sure we can connect to it and that the
        necessary indexes are set up.
        """
        if not doc.get('db', ''):
            raise ValidationException('Database name must not be empty.', 'db')
        if '.' in doc['db'] or ' ' in doc['db']:
            raise ValidationException('Database name cannot contain spaces'
                                      ' or periods.', 'db')

        try:
            chunkColl = getDbConnection(
                doc.get('mongohost'), doc.get('replicaset'), autoRetry=False,
                serverSelectionTimeoutMS=10000)[doc['db']].chunk
            _ensureChunkIndices(chunkColl)
        except pymongo.errors.ServerSelectionTimeoutError as e:
            raise ValidationException(
                'Could not connect to the database: %s' % str(e))

        return doc

    @staticmethod
    def fileIndexFields():
        return ['sha512', 'chunkUuid']

    def __init__(self, assetstore):
        """
        :param assetstore: The assetstore to act on.
        """
        super(GridFsAssetstoreAdapter, self).__init__(assetstore)
        recent = False
        try:
            # Guard in case the connectionArgs is unhashable
            key = (self.assetstore.get('mongohost'),
                   self.assetstore.get('replicaset'))
            if key in _recentConnections:
                recent = (time.time() - _recentConnections[key]['created']
                          < RECENT_CONNECTION_CACHE_TIME)
        except TypeError:
            key = None
        try:
            # MongoClient automatically reuses connections from a pool, but we
            # want to avoid redoing ensureChunkIndices each time we get such a
            # connection.
            client = getDbConnection(self.assetstore.get('mongohost'),
                                     self.assetstore.get('replicaset'),
                                     quiet=recent)
            self.chunkColl = MongoProxy(client[self.assetstore['db']].chunk)
            if not recent:
                _ensureChunkIndices(self.chunkColl)
                if key is not None:
                    if len(_recentConnections) >= RECENT_CONNECTION_CACHE_MAX_SIZE:
                        _recentConnections.clear()
                    _recentConnections[key] = {
                        'created': time.time()
                    }
        except pymongo.errors.ConnectionFailure:
            logger.error('Failed to connect to GridFS assetstore %s',
                         self.assetstore['db'])
            self.chunkColl = 'Failed to connect'
            self.unavailable = True
        except pymongo.errors.ConfigurationError:
            logger.exception('Failed to configure GridFS assetstore %s',
                             self.assetstore['db'])
            self.chunkColl = 'Failed to configure'
            self.unavailable = True

    def initUpload(self, upload):
        """
        Creates a UUID that will be used to uniquely link each chunk to
        """
        upload['chunkUuid'] = uuid.uuid4().hex
        upload['sha512state'] = _hash_state.serializeHex(sha512())
        return upload

    def uploadChunk(self, upload, chunk):
        """
        Stores the uploaded chunk in fixed-sized pieces in the chunks
        collection of this assetstore's database.
        """
        # If we know the chunk size is too large or small, fail early.
        self.checkUploadSize(upload, self.getChunkSize(chunk))

        if isinstance(chunk, six.text_type):
            chunk = chunk.encode('utf8')

        if isinstance(chunk, six.binary_type):
            chunk = BytesIO(chunk)

        # Restore the internal state of the streaming SHA-512 checksum
        checksum = _hash_state.restoreHex(upload['sha512state'], 'sha512')

        # TODO: when saving uploads is optional, we can conditionally try to
        # fetch the last chunk.  Add these line before `lastChunk = ...`:
        #   lastChunk = None
        #   if '_id' in upload or upload['received'] != 0:
        lastChunk = self.chunkColl.find_one({
            'uuid': upload['chunkUuid']
        }, projection=['n'], sort=[('n', pymongo.DESCENDING)])
        if lastChunk:
            # This bit of code will only do anything if there is a discrepancy
            # between the received count of the upload record and the length of
            # the file stored as chunks in the database. This code updates the
            # sha512 state with the difference before reading the bytes sent
            # from the user.
            if self.requestOffset(upload) > upload['received']:
                # This isn't right -- the last received amount may not be a
                # complete chunk.
                cursor = self.chunkColl.find({
                    'uuid': upload['chunkUuid'],
                    'n': {'$gte': upload['received'] // CHUNK_SIZE}
                }, projection=['data']).sort('n', pymongo.ASCENDING)
                for result in cursor:
                    checksum.update(result['data'])
        n = lastChunk['n'] + 1 if lastChunk else 0

        size = 0
        startingN = n

        while upload['received'] + size < upload['size']:
            data = chunk.read(CHUNK_SIZE)
            if not data:
                break
            # If a timeout occurs while we are trying to load data, we might
            # have succeeded, in which case we will get a DuplicateKeyError
            # when it automatically retries.  Therefore, log this error but
            # don't stop.
            try:
                self.chunkColl.insert_one({
                    'n': n,
                    'uuid': upload['chunkUuid'],
                    'data': bson.binary.Binary(data)
                })
            except pymongo.errors.DuplicateKeyError:
                logger.info('Received a DuplicateKeyError while uploading, '
                            'probably because we reconnected to the database '
                            '(chunk uuid %s part %d)', upload['chunkUuid'], n)
            n += 1
            size += len(data)
            checksum.update(data)
        chunk.close()

        try:
            self.checkUploadSize(upload, size)
        except ValidationException:
            # The user tried to upload too much or too little.  Delete
            # everything we added
            self.chunkColl.delete_many({
                'uuid': upload['chunkUuid'],
                'n': {'$gte': startingN}
            })
            raise

        # Persist the internal state of the checksum
        upload['sha512state'] = _hash_state.serializeHex(checksum)
        upload['received'] += size
        return upload

    def requestOffset(self, upload):
        """
        The offset will be the CHUNK_SIZE * total number of chunks in the
        database for this file. We return the max of that and the received
        count because in testing mode we are uploading chunks that are smaller
        than the CHUNK_SIZE, which in practice will not work.
        """
        lastChunk = self.chunkColl.find_one({
            'uuid': upload['chunkUuid']
        }, projection=['n'], sort=[('n', pymongo.DESCENDING)])

        if lastChunk is None:
            offset = 0
        else:
            offset = lastChunk['n'] * CHUNK_SIZE
        return max(offset, upload['received'])

    def finalizeUpload(self, upload, file):
        """
        Grab the final state of the checksum and set it on the file object,
        and write the generated UUID into the file itself.
        """
        hash = _hash_state.restoreHex(upload['sha512state'], 'sha512').hexdigest()

        file['sha512'] = hash
        file['chunkUuid'] = upload['chunkUuid']
        file['chunkSize'] = CHUNK_SIZE

        return file

    def downloadFile(self, file, offset=0, headers=True, endByte=None,
                     contentDisposition=None, extraParameters=None, **kwargs):
        """
        Returns a generator function that will be used to stream the file from
        the database to the response.
        """
        if endByte is None or endByte > file['size']:
            endByte = file['size']

        if headers:
            setResponseHeader('Accept-Ranges', 'bytes')
            self.setContentHeaders(file, offset, endByte, contentDisposition)

        # If the file is empty, we stop here
        if endByte - offset <= 0:
            return lambda: ''

        n = 0
        chunkOffset = 0

        # We must "seek" to the correct chunk index and local offset
        if offset > 0:
            n = offset // file['chunkSize']
            chunkOffset = offset % file['chunkSize']

        cursor = self.chunkColl.find({
            'uuid': file['chunkUuid'],
            'n': {'$gte': n}
        }, projection=['data']).sort('n', pymongo.ASCENDING)

        def stream():
            co = chunkOffset  # Can't assign to outer scope without "nonlocal"
            position = offset
            shouldBreak = False

            for chunk in cursor:
                chunkLen = len(chunk['data'])

                if position + chunkLen - co > endByte:
                    chunkLen = endByte - position + co
                    shouldBreak = True

                yield chunk['data'][co:chunkLen]

                if shouldBreak:
                    break

                position += chunkLen - co

                if co > 0:
                    co = 0

        return stream

    def deleteFile(self, file):
        """
        Delete all of the chunks in the collection that correspond to the
        given file.
        """
        q = {
            'chunkUuid': file['chunkUuid'],
            'assetstoreId': self.assetstore['_id']
        }
        matching = File().find(q, limit=2, projection=[])
        if matching.count(True) == 1:
            # If we can't reach the database, we return anyway.  A system check
            # will be necessary to remove the abandoned file.  Since we already
            # can handle that case, tell Mongo to use a 0 write concern -- we
            # don't need to know that the chunks have been deleted, and this
            # can be faster.
            try:
                self.chunkColl.with_options(
                    write_concern=pymongo.WriteConcern(w=0)).delete_many(
                        {'uuid': file['chunkUuid']})
            except pymongo.errors.AutoReconnect:
                pass

    def cancelUpload(self, upload):
        """
        Delete all of the chunks associated with a given upload.
        """
        self.chunkColl.delete_many({'uuid': upload['chunkUuid']})



import json
import os

from ansible.module_utils.basic import AnsibleModule
from inspect import getmembers, ismethod, isfunction, getargspec
import requests
import six

try:
    from girder_client import GirderClient, AuthenticationError
    HAS_GIRDER_CLIENT = True
except ImportError:
    HAS_GIRDER_CLIENT = False

try:
    from girder.utility.s3_assetstore_adapter import DEFAULT_REGION
except ImportError:
    DEFAULT_REGION = 'us-east-1'

__version__ = '0.3.0'

DOCUMENTATION = """
---
module: girder
author: "Kitware, Inc. <kitware@kitware.com>
version_added: "0.1"
short_description: A module that wraps girder_client
requirements: [ girder_client==1.1.0 ]
description:
   - Manage a girder instance using the RESTful API
options:
    host:
        required: false
        default: 'localhost'
        description:
            - domain or IP of the host running girder
    port:
        required: false
        default: '80' for http, '443' for https
        description:
            - port the girder instance is running on

    apiRoot:
        required: false
        default: '/api/v1'
        description:
            - path on server corresponding to the root of Girder REST API

    apiUrl:
        required: false
        default: None
        description:
            - full URL base of the girder instance API
    apiKey:
        required: false
        default: None
        description:
            - pass in an apiKey instead of username/password


    scheme:
        required: false
        default: 'http'
        description:
            - A string containing the scheme for the Girder host

    dryrun:
        required: false
        default: None (passed through)
        description:
            - See GirderClient.__init__()

    blacklist:
        required: false
        default: None (passed through)
        description:
            - See GirderClient.__init__()

    username:
        required: true
        description:
            - Valid username for the system
            - Required with password
            - must be specified if 'token' is not specified
                - (See note on 'user')

    password:
        required: true
        description:
            - Valid password for the system
            - Required with username
            - must be specified if 'token' is not specified
                - (See note on 'user')
    token:
        required: true
        description:
            - A girder client token
            - Can be retrieved by accessing the accessing the 'token' attribute
              from a successfully authenticated call to girder in a previous
              task.
            - Required if 'username' and 'password' are not specified
                - (See note on 'user')
    state:
        required: false
        default: "present"
        choices: ["present", "absent"]
        description:
            - Used to indicate the presence or absence of a resource
              - e.g.,  user, assetstore

    user:
        required: false
        description:
            - If using the 'user' task, you are NOT REQUIRED to pass in a
              'username' & 'password',  or a 'token' attributes. This is because
              the first user created on an fresh install of girder is
              automatically made an administrative user. Once you are certain
              you have an admin user you should use those credentials in all
              subsequent tasks that use the 'user' task.

            - Takes a mapping of key value pairs
              options:
                  login:
                      required: true
                      description:
                          - The login name of the user
                  password:
                      required: true
                      description:
                          - The password of the user

                  firstName:
                      required: false
                      default: pass through to girder client
                      description:
                          - The first name of the user

                  lastName:
                      required: false
                      default: pass through to girder client
                      description:
                          - The last name of the user
                  email:
                      required: false
                      default: pass through to girder client
                      description:
                          - The email of the user
                  admin:
                      required: false
                      default: false
                      description:
                          - If true,  make the user an administrator.


    assetstore:
        required: false
        description:
            - Specifies an assetstore
            - Takes many options depending on 'type'
              options:
                  name:
                      required: true
                      description:
                          - Name of the assetstore
                  type:
                      required: true
                      choices: ['filesystem', 'gridfs', 's3', 'hdfs', 'database']
                      description:
                          - Currently only 'filesystem' has been tested
                  readOnly:
                      required: false
                      default: false
                      description:
                          - Should the assetstore be read only?
                  current:
                      required: false
                      default: false
                      description:
                          - Should the assetstore be set as the current
                            assetstore?

              options (filesystem):
                  root:
                      required: true
                      description:
                          -  Filesystem path to the assetstore

              options (gridfs) (EXPERIMENTAL):
                   db:
                       required: true
                       description:
                           - database name
                   mongohost:
                       required: true
                       description:
                           - Mongo host URI

                   replicaset:
                       required: false
                       default: ''
                       description:
                           - Replica set name

              options (s3) (EXPERIMENTAL):
                   bucket:
                       required: true
                       description:
                           - The S3 bucket to store data in

                   prefix:
                       required: false
                       description:
                           - Optional path prefix within the bucket under which
                             files will be stored

                   accessKeyId:
                       required: false
                       description:
                           - the AWS access key ID to use for authentication

                   secret:
                       required: false
                       description:
                           - the AWS secret key to use for authentication

                   service:
                       required: false
                       default: ''
                       description:
                           - The S3 service host (for S3 type)
                           - Only set this if you're not using AWS S3.
                           - This can be used to specify a protocol and port
                             -  use the form [http[s]://](host domain)[:(port)]
                           - Do not include the bucket name here

                   region:
                       required: false
                       default: us-east

                   inferCredentials:
                       required: false
                       default: false

              options (hdfs) (EXPERIMENTAL):
                   host:
                       required: true
                       description:
                           - None
                   port:
                       required: true
                       description:
                           - None
                   path:
                       required: true
                       description:
                           - None
                   user:
                       required: true
                       description:
                           - None
                   webHdfsPort
                       required: true
                       description:
                           - None

    group:
        required: false
        description:
            - Create a group with pre-existing users
        options:
            name:
                required: true
                description:
                    - Name of the group

            description:
                required: false
                description:
                    - Description of the group
            users:
                required: false
                type: list
                description:
                    - List of dicts with users login and their level
                options:
                    login:
                        required: true
                        description:
                            - the login name
                    type:
                        required: true
                        choices: ["member", "moderator", "admin"]
                        description:
                            - Access level for that user in the group

    collection:
        required: false
        description:
            - Create a collection
        options:
            name:
                required: true
                description:
                    - Name of the collection

            description:
                required: false
                description:
                    - Description of the collection

            folders:
                required: false
                description:
                    - A list of folder options
                    - Specified by the 'folder' option to the girder module
                    - (see 'folder:')
            public:
                required: false
                description:
                    - Set to true if the collection is public or false if private
            access:
                required: false
                description:
                    - Set the access for the collection/folder
                options:
                    users:
                        required: false
                        description:
                            - list of login/type arguments
                            - login is a user login
                            - type is one of 'admin', 'moderator', 'member'
                    groups:
                        required: false
                        description:
                            - list of name/type arguments
                            - name is a group name
                            - type is one of 'admin', 'moderator', 'member'

    folder:
        required: false
        description:
            - Create a folder
        options:
            name:
                required: true
                description:
                    - Name of the folder

            description:
                required: false
                description:
                    - Description of the folder
            parentType:
                required: true
                choices: ["user", "folder", "collection"]
                description:
                    - The type of the parent
            parentId:
                required: true
                description:
                    - The ID of the parent collection/folder/user
            folders:
                required: false
                description:
                    - A list of folder options
                    - Specified by the 'folder' option to the girder module
                    - (see 'folder:')
            public:
                required: false
                description:
                    - Set to true if the folder is public or false if private
            access:
                required: false
                description:
                    - Set the access for the collection/folder
                options:
                    users:
                        required: false
                        description:
                            - list of login/type arguments
                            - login is a user login
                            - type is one of 'admin', 'moderator', 'member'
                    groups:
                        required: false
                        description:
                            - list of name/type arguments
                            - name is a group name
                            - type is one of 'admin', 'moderator', 'member'

    item:
        required: false
        description:
            - Create a item
        options:
            name:
                required: true
                description:
                    - Name of the item

            description:
                required: false
                description:
                    - Description of the item
            folderId:
                required: true
                description:
                    - The ID of the parent collection

     files:
        required: false
        description:
            - Uploads a list of files to an item
        options:
            itemId:
                required: true
                description:
                    - the parent item for the file
            sources:
                required: true
                description:
                    - list of local file paths
                    - files will be uploaded to the item

     setting:
        required: false
        description:
            - Get/set the values of system settings
        options:
            key:
                required: true
                description:
                    - The key identifying this setting
            value:
                required: true if state = present, else false
                description:
                    - The value to set
"""

EXAMPLES = """


#############
# Example using 'user'
###


# Ensure "admin" user exists
- name: Create 'admin' User
  girder:
    user:
      firstName: "John"
      lastName: "Doe"
      login: "admin"
      password: "letmein"
      email: "john.due@test.com"
      admin: yes
    state: present

# Ensure a 'foobar' user exists
- name: Create 'foobar' User
  girder:
    username: "admin"
    password: "letmein"
    user:
      firstName: "Foo"
      lastName: "Bar"
      login: "foobar"
      password: "foobarbaz"
      email: "foo.bar@test.com"
      admin: yes
    state: present

# Remove the 'foobar' user
- name: Remove 'foobar' User
  username: "admin"
  password: "letmein"
  girder:
    user:
      login: "foobar"
      password: "foobarbaz"
    state: absent

############
# Examples using Group
#

# Create an 'alice' user
- name: Create 'alice' User
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    user:
      firstName: "Alice"
      lastName: "Doe"
      login: "alice"
      password: "letmein"
      email: "alice.doe@test.com"
    state: present

# Create a 'bill' user
- name: Create 'bill' User
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    user:
      firstName: "Bill"
      lastName: "Doe"
      login: "bill"
      password: "letmein"
      email: "bill.doe@test.com"
    state: present

# Create a 'chris' user
- name: Create 'chris' User
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    user:
      firstName: "Chris"
      lastName: "Doe"
      login: "chris"
      password: "letmein"
      email: "chris.doe@test.com"
    state: present

- name: Create a test group with users
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    group:
      name: "Test Group"
      description: "Basic test group"
      users:
        - login: alice
          type: member
        - login: bill
          type: moderator
        - login: chris
          type: admin
    state: present

# Remove Bill from the group,
# Note that 'group' list is idempotent - it describes the desired state

- name: Remove bill from group
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    group:
      name: "Test Group"
      description: "Basic test group"
      users:
        - login: alice
          type: member
        - login: chris
          type: admin
    state: present


############
# Filesystem Assetstore Tests
#

- name: Create filesystem assetstore
  girder:
    username: "admin"
     password: "letmein"
     assetstore:
       name: "Temp Filesystem Assetstore"
       type: "filesystem"
       root: "/data/"
       current: true
     state: present

- name: Delete filesystem assetstore
  girder:
    username: "admin"
    password: "letmein"
    assetstore:
      name: "Temp Filesystem Assetstore"
      type: "filesystem"
      root: "/tmp/"
    state: absent


############
# Examples using collections, folders, items and files
#

# Creates a test collection called "Test Collection"
- name: Create collection
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    collection:
      name: "Test Collection"
      description: "A test collection"
  register: test_collection

# Creates a folder called "test folder" under "Test Collection"
- name: Create folder
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    folder:
      parentType: "collection"
      parentId: "{{test_collection['gc_return']['_id'] }}"
      name: "test folder"
      description: "A test folder"
  register: test_folder

# Creates an item called "test item" under "test folder"
- name: Create an item
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    item:
      folderId: "{{test_folder['gc_return']['_id'] }}"
      name: "test item"
      description: "A test item"
  register: test_item

# Upload files on the localhost at /tmp/data/test1.txt and
# /tmp/data/test2.txt to the girder instance under the item
# "test item"
# Note:  the list is idempotent and will remove files that are
# not listed under the item. Files are checked for both name
# and size to determine if they should be updated.
- name: Upload files
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    files:
      itemId: "{{ test_item['gc_return']['_id'] }}"
      sources:
        - /tmp/data/test1.txt
        - /tmp/data/test2.txt
  register: retval


############
# Examples Using collection/folder hierarchy
#

- name: Create collection with a folder and a subfolder
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    collection:
      name: "Test Collection"
      description: "A test collection"
      folders:
        - name: "test folder"
          description: "A test folder"
          folders:
            - name: "test subfolder"
            - name: "test subfolder 2"
  register: test_collection



############
# Examples Setting access to files/folders
#


- name: Create collection with access
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    collection:
      name: "Test Collection"
      description: "A test collection"
      public: no
      access:
        users:
          - login: alice
            type: admin
          - login: chris
            type: member
  register: test_collection


- name: Add group to Test Collection
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    collection:
      name: "Test Collection"
      description: "A test collection"
      public: no
      access:
        users:
          - login: alice
            type: admin
          - login: bill
            type: moderator
          - login: chris
            type: member
        groups:
          - name: Test Group
            type: member
  register: test_collection

- name: Add Test Folder with access
  girder:
    port: 8080
    username: "admin"
    password: "letmein"
    folder:
      parentType: "collection"
      parentId: "{{test_collection['gc_return']['_id'] }}"
      name: "test folder"
      description: "A test folder"
      access:
        users:
          - login: bill
            type: admin
        groups:
          - name: Test Group
            type: member
  register: test_folder



############
# Examples using get
#


# Get my info
- name: Get users from http://localhost:80/api/v1/users
  girder:
    username: 'admin'
    password: 'letmein'
    get:
      path: "users"
    register: ret_val

# Prints debugging messages with the emails of the users
# From the last task by accessing 'gc_return' of the registered
# variable 'ret_val'
- name: print emails of users
  debug: msg="{{ item['email'] }}"
  with_items: "{{ ret_val['gc_return'] }}"


#############
# Advanced usage
#

# Supports get, post, put, delete methods,  but does
# not guarantee idempotence on these methods!

- name: Run a system check
  girder:
    username: "admin"
    password: "letmein"
    put:
      path: "system/check"

# An example of posting an item to Girder
# Note that this is NOT idempotent. Running
# multiple times will create "An Item", "An Item (1)",
# "An Item (2)", etc..

- name: Get Me
  girder:
    username: "admin"
    password: "letmein"
    get:
      path: "user/me"
  register: ret

# Show use of 'token' for subsequent authentication
- name: Get my public folder
  girder:
    token: "{{ ret['token'] }}"
    get:
      path: "folder"
      parameters:
        parentType: "user"
        parentId: "{{ ret['gc_return']['_id'] }}"
        text: "Public"
  register: ret


- name: Post an item to my public folder
  girder:
    host: "data.kitware.com"
    scheme: 'https'
    token: "{{ ret['token'] }}"
    post:
      path: "item"
      parameters:
        folderId: "{{ ret['gc_return'][0]['_id'] }}"
        name: "An Item"


"""


def unjsonify(a):
    """Convert json parts to python objects.

    Tries to convert a json string or a compund object consisting of partial
    json strings to a full python object.  Returns either a json scalar (string,
    int, float), a list or dict of json scalars, or any combination of lists and
    dicts that eventually end at json scalars.

    Note that this function does not detect cycles in an object graph and, if
    provided an object with one, will run until out of memory, at the stack
    limit, or until at some other run-time limit.
    """
    # if string, try to loads() it
    if isinstance(a, six.string_types):
        try:
            a = json.loads(a)
            # pass-through to below
        except ValueError:
            return a

    if isinstance(a, list):
        return [unjsonify(x) for x in a]

    if isinstance(a, dict):
        return {str(k): unjsonify(v) for k, v in a.items()}

    return None


def class_spec(cls, include=None):
    include = include if include is not None else []

    for fn, method in getmembers(cls, predicate=lambda f: ismethod(f) or isfunction(f)):
        if fn in include:
            spec = getargspec(method)
            # Note: must specify the kind of data we accept
            #       In all most all cases this will be a dict
            #       where variable names become keys used in yaml
            #       but if we have a vararg then we need to set
            #       this to a list.
            kind = 'dict' if spec.varargs is None else 'list'

            # spec.args[1:] so we don't include 'self'
            params = spec.args[1:]
            d = len(spec.defaults) if spec.defaults is not None else 0
            r = len(params) - d

            yield (fn, {'required': params[:r],
                        'optional': params[r:],
                        'type': kind})


class Resource(object):
    known_resources = ['collection', 'folder', 'item', 'group']

    def __init__(self, client, resource_type):
        self._resources = None
        self._resources_by_name = None
        self.client = client

        if resource_type in self.known_resources:
            self.resource_type = resource_type
        else:
            raise Exception('{} is an unknown resource!'.format(resource_type))

    @property
    def resources(self):
        if self._resources is None:
            self._resources = {r['_id']: r for r
                               in self.client.get(self.resource_type)}
        return self._resources

    @property
    def resources_by_name(self):
        if self._resources_by_name is None:
            self._resources_by_name = {r['name']: r
                                       for r in self.resources.values()}

        return self._resources_by_name

    def __apply(self, _id, func, *args, **kwargs):
        if _id in self.resources.keys():
            ret = func('{}/{}'.format(self.resource_type, _id),
                       *args, **kwargs)
            self.client.changed = True
        return ret

    def id_exists(self, _id):
        return _id in self.resources.keys()

    def name_exists(self, _name):
        return _name in self.resources_by_name.keys()

    def create(self, body, **kwargs):
        try:
            ret = self.client.post(self.resource_type, body, **kwargs)
            self.client.changed = True
        except requests.HTTPError as htErr:
            try:
                # If we can't create the item,  try and return
                # The item with the same name
                ret = self.resources_by_name[kwargs['name']]
            except KeyError:
                raise htErr
        return ret

    def read(self, _id):
        return self.resources[_id]

    def read_by_name(self, name):
        return self.resources_by_name[name]['_id']

    def update(self, _id, body, **kwargs):
        if _id in self.resources:
            current = self.resources[_id]
            # if body is a subset of current we don't actually need to update
            try:
                if set(body.items()) <= set(
                        {k: v for k, v in current.items() if k in body}.items()):
                    return current
            except TypeError:
                # If a current value is unhashable, we throw a TypeError, but
                # should still update it.
                pass
            return self.__apply(_id, self.client.put, body, **kwargs)
        else:
            raise Exception('{} does not exist!'.format(_id))

    def update_by_name(self, name, body, **kwargs):
        return self.update(self.resources_by_name[name]['_id'],
                           body, **kwargs)

    def delete(self, _id):
        return self.__apply(_id, self.client.delete)

    def delete_by_name(self, name):
        try:
            return self.delete(self.resources_by_name[name]['_id'])
        except KeyError:
            return {}


class AccessMixin(object):

    def get_access(self, _id):
        return self.client.get('{}/{}/access'
                               .format(self.resource_type, _id))

    def put_access(self, _id, access, public=True):
        current_access = self.get_access(_id)

        if set([tuple(u.values()) for u in access['users']]) ^ \
           set([(u['id'], u['level']) for u in current_access['users']]):
            self.client.changed = True

        if set([tuple(g.values()) for g in access['groups']]) ^ \
           set([(u['id'], u['level']) for u in current_access['groups']]):
            self.client.changed = True

        return self.client.put('{}/{}/access'
                               .format(self.resource_type, _id),
                               dict(access=json.dumps(access),
                                    public='true' if public else 'false'))


class CollectionResource(AccessMixin, Resource):
    def __init__(self, client):
        super(CollectionResource, self).__init__(client, 'collection')


class GroupResource(Resource):
    def __init__(self, client):
        super(GroupResource, self).__init__(client, 'group')


class FolderResource(AccessMixin, Resource):
    def __init__(self, client, parentType, parentId):
        super(FolderResource, self).__init__(client, 'folder')
        self.parentType = parentType
        self.parentId = parentId

    @property
    def resources(self):
        if self._resources is None:
            self._resources = {r['_id']: r for r
                               in self.client.get(self.resource_type, {
                                   'parentType': self.parentType,
                                   'parentId': self.parentId
                               })}
            # parentType is stored as parrentCollection in database
            # We need parentType to be available so we can do set
            # comparison to check if we are updating parentType (e.g.
            # Moving a subfolder from a folder to a collection)
            for _id in self._resources.keys():
                self._resources[_id]['parentType'] = \
                    self._resources[_id]['parentCollection']
        return self._resources


class ItemResource(Resource):
    def __init__(self, client, folderId):
        super(ItemResource, self).__init__(client, 'item')
        self.folderId = folderId

    @property
    def resources(self):
        if self._resources is None:
            self._resources = {r['_id']: r for r
                               in self.client.get(self.resource_type, {
                                   'folderId': self.folderId
                               })}
        return self._resources


class GirderClientModule(GirderClient):

    # Exclude these methods from both 'raw' mode
    _include_methods = ['get', 'put', 'post', 'delete', 'patch',
                        'user', 'assetstore',
                        'collection', 'folder', 'item', 'files',
                        'group', 'setting']

    _debug = True

    def exit(self):
        if not self._debug:
            del self.message['debug']

        self.module.exit_json(changed=self.changed, **self.message)

    def fail(self, msg):
        self.module.fail_json(msg=msg)

    def __init__(self):
        self.changed = False
        self.message = {'msg': 'Success!', 'debug': {}}

        self.spec = dict(class_spec(self.__class__,
                                    GirderClientModule._include_methods))
        self.required_one_of = self.spec.keys()

        # Note: if additional types are added o girder this will
        # have to be updated!
        self.access_types = {'member': 0, 'moderator': 1, 'admin': 2}

    def __call__(self, module):
        self.module = module

        super(GirderClientModule, self).__init__(
            **{p: self.module.params[p] for p in
               ['host', 'port', 'apiRoot', 'apiUrl',
                'scheme', 'dryrun', 'blacklist']
               if module.params[p] is not None})
        # If a username and password are set
        if self.module.params['username'] is not None:
            try:
                self.authenticate(
                    username=self.module.params['username'],
                    password=self.module.params['password'])

            except AuthenticationError:
                self.fail('Could not Authenticate!')

        elif self.module.params['apiKey'] is not None:
            try:
                self.authenticate(
                    apiKey=self.module.params['apiKey'])

            except AuthenticationError:
                self.fail('Could not Authenticate!')

        # If a token is set
        elif self.module.params['token'] is not None:
            self.token = self.module.params['token']

        # Else error if we're not trying to create a user
        elif self.module.params['user'] is None:
            self.fail('Must pass in either username & password, '
                      'or a valid girder_client token')

        self.message['token'] = self.token

        for method in self.required_one_of:
            if self.module.params[method] is not None:
                self.__process(method)
                self.exit()

        self.fail('Could not find executable method!')

    def __process(self, method):
        # Parameters from the YAML file
        params = self.module.params[method]
        # Final list of arguments to the function
        args = []
        # Final list of keyword arguments to the function
        kwargs = {}

        if isinstance(params, dict):
            for arg_name in self.spec[method]['required']:
                if arg_name not in params.keys():
                    self.fail('%s is required for %s' % (arg_name, method))
                args.append(params[arg_name])

            for kwarg_name in self.spec[method]['optional']:
                if kwarg_name in params.keys():
                    kwargs[kwarg_name] = params[kwarg_name]

        elif isinstance(params, list):
            args = params
        else:
            args = [params]

        ret = getattr(self, method)(*args, **kwargs)

        self.message['debug']['method'] = method
        self.message['debug']['args'] = args
        self.message['debug']['kwargs'] = kwargs
        self.message['debug']['params'] = params

        self.message['gc_return'] = ret

    def files(self, itemId, sources=None):
        ret = {'added': [],
               'removed': []}

        files = self.get('item/{}/files'.format(itemId))

        if self.module.params['state'] == 'present':

            file_dict = {f['name']: f for f in files}

            source_dict = {os.path.basename(s): {
                'path': s,
                'name': os.path.basename(s),
                'size': os.path.getsize(s)} for s in sources}

            source_names = set([(s['name'], s['size'])
                                for s in source_dict.values()])

            file_names = set([(f['name'], f['size'])
                              for f in file_dict.values()])

            for n, _ in (file_names - source_names):
                self.delete('file/{}'.format(file_dict[n]['_id']))
                ret['removed'].append(file_dict[n])

            for n, _ in (source_names - file_names):
                self.uploadFileToItem(itemId, source_dict[n]['path'])
                ret['added'].append(source_dict[n])

        elif self.module.params['state'] == 'absent':
            for f in files:
                self.delete('file/{}'.format(f['_id']))
                ret['removed'].append(f)

        if len(ret['added']) != 0 or len(ret['removed']) != 0:
            self.changed = True

        return ret

    def _get_user_by_login(self, login):
        try:
            user = self.get('/resource/lookup',
                            {'path': '/user/{}'.format(login)})
        except requests.HTTPError:
            user = None
        return user

    def _get_group_by_name(self, name):
        try:
            # Could potentially fail if we have more 50 groups
            group = {g['name']: g for g in self.get('group')}['name']
        except (KeyError, requests.HTTPError):
            group = None
        return group

    def group(self, name, description, users=None, debug=False):

        r = GroupResource(self)
        valid_fields = [('name', name),
                        ('description', description)]

        if self.module.params['state'] == 'present':
            if r.name_exists(name):
                ret = r.update_by_name(name, {k: v for k, v in valid_fields
                                              if v is not None})
            else:
                ret = r.create({k: v for k, v in valid_fields
                                if v is not None})

            if users is not None:
                ret['added'] = []
                ret['removed'] = []
                ret['updated'] = []

                group_id = ret['_id']

                # Validate and normalize the user list
                for user in users:
                    assert 'login' in user.keys(), \
                        'User list must have a login attribute'

                    user['type'] = self.access_types.get(
                        user.get('type', 'member'), 'member')

                # dict of passed in login -> type
                user_levels = {u['login']: u['type'] for u in users}

                # dict of current login -> user information for this group
                members = {m['login']: m for m in
                           self.get('group/{}/member'.format(group_id))}

                # Add these users
                for login in (set(user_levels.keys()) - set(members.keys())):
                    user = self._get_user_by_login(login)
                    if user is not None:
                        # add user at level
                        self.post('group/{}/invitation'.format(group_id),
                                  {'userId': user['_id'],
                                   'level': user_levels[login],
                                   'quiet': True,
                                   'force': True})
                        ret['added'].append(user)
                    else:
                        raise Exception('{} is not a valid login!'
                                        .format(login))

                # Remove these users
                for login in (set(members.keys()) - set(user_levels.keys())):
                    self.delete('/group/{}/member'.format(group_id),
                                {'userId': members[login]['_id']})
                    ret['removed'].append(members[login])

                # Set of users that potentially need to be updated
                if len(set(members.keys()) & set(user_levels.keys())):
                    group_access = self.get('group/{}/access'.format(group_id))
                    # dict of current login -> access information for this group
                    user_access = {m['login']: m
                                   for m in group_access['access']['users']}

                    # dict of login -> level for the current group
                    # Note:
                    #  Here we join members with user_access - if the member
                    #  is not in user_access then the member has a level of 0 by
                    #  default. This gives us  a complete list of every login,
                    #  and its access level, including those that are IN the
                    #  group, but have no permissions ON the group.
                    member_levels = {m['login']:
                                     user_access.get(m['login'],
                                                     {'level': 0})['level']
                                     for m in members.values()}

                    ret = self._promote_or_demote_in_group(ret,
                                                           member_levels,
                                                           user_levels,
                                                           group_id)

                # Make sure 'changed' is handled correctly if we've
                # manipulated the group's users in any way
                if (len(ret['added']) != 0 or len(ret['removed']) != 0
                        or len(ret['updated']) != 0):
                    self.changed = True

        elif self.module.params['state'] == 'absent':
            ret = r.delete_by_name(name)

        return ret

    def _promote_or_demote_in_group(self, ret, member_levels, user_levels,
                                    group_id):
        """Promote or demote a set of users.

        :param ret: the current dict of return values
        :param members_levels: the current access levels of each member
        :param user_levels: the desired levels of each member
        :param types: a mapping between resource names and access levels
        :returns: info about what has (or has not) been updated
        :rtype: dict

        """
        reverse_type = {v: k for k, v in self.access_types.items()}

        for login in (set(member_levels.keys())
                      & set(user_levels.keys())):
            user = self._get_user_by_login(login)
            _id = user['_id']

            # We're promoting
            if member_levels[login] < user_levels[login]:
                resource = reverse_type[user_levels[login]]
                self.post('group/{}/{}'
                          .format(group_id, resource),
                          {'userId': _id})

                user['from_level'] = member_levels[login]
                user['to_level'] = user_levels[login]
                ret['updated'].append(user)

            # We're demoting
            elif member_levels[login] > user_levels[login]:
                resource = reverse_type[member_levels[login]]
                self.delete('group/{}/{}'
                            .format(group_id, resource),
                            {'userId': _id})

                # In case we're not demoting to member make sure
                # to update to promote to whatever level we ARE
                # demoting too now that our user is a only a member
                if user_levels[login] != 0:
                    resource = reverse_type[user_levels[login]]
                    self.post('group/{}/{}'
                              .format(group_id, resource),
                              {'userId': _id})

                    user['from_level'] = member_levels[login]
                    user['to_level'] = user_levels[login]
                    ret['updated'].append(user)

        return ret

    def item(self, name, folderId, description=None, files=None,
             access=None, debug=False):
        ret = {}
        r = ItemResource(self, folderId)
        valid_fields = [('name', name),
                        ('description', description),
                        ('folderId', folderId)]

        if self.module.params['state'] == 'present':
            if r.name_exists(name):
                ret = r.update_by_name(name, {k: v for k, v in valid_fields
                                              if v is not None})
            else:
                ret = r.create({k: v for k, v in valid_fields
                                if v is not None})
        # handle files here

        elif self.module.params['state'] == 'absent':
            ret = r.delete_by_name(name)

        return ret

    def folder(self, name, parentId, parentType, description=None,
               public=True, folders=None, access=None, debug=False):

        ret = {}

        assert parentType in ['collection', 'folder', 'user'], \
            'parentType must be collection or folder'

        r = FolderResource(self, parentType, parentId)
        valid_fields = [('name', name),
                        ('description', description),
                        ('parentType', parentType),
                        ('parentId', parentId)]

        if self.module.params['state'] == 'present':
            if r.name_exists(name):
                ret = r.update_by_name(name, {k: v for k, v in valid_fields
                                              if v is not None})
            else:
                valid_fields = valid_fields + [('public', public)]
                ret = r.create({k: v for k, v in valid_fields
                                if v is not None})

            if folders is not None:
                self._process_folders(folders, ret['_id'], 'folder')

            # handle access here
            if access is not None:
                _id = ret['_id']
                ret['access'] = self._access(r, access, _id, public=public)

        elif self.module.params['state'] == 'absent':
            ret = r.delete_by_name(name)

        return ret

    def _access(self, r, access, _id, public=True):
        access_list = {'users': [], 'groups': []}
        users = access.get('users', None)
        groups = access.get('groups', None)

        if groups is not None:
            assert set(g['type'] for g in groups if 'type' in g) <= \
                set(self.access_types.keys()), 'Invalid access type!'

            # Hash of name -> group information
            # used to get user id's for access control lists
            all_groups = {g['name']: g for g in self.get('group')}

            access_list['groups'] = [{'id': all_groups[g['name']]['_id'],
                                      'level': self.access_types[g['type']]
                                      if 'type' in g else g['level']}
                                     for g in groups]

        if users is not None:

            assert set(u['type'] for u in users if 'type' in u) <= \
                set(self.access_types.keys()), 'Invalid access type!'

            # Hash of login -> user information
            # used to get user id's for access control lists
            current_users = {u['login']: self._get_user_by_login(u['login'])
                             for u in users}

            access_list['users'] = [{'id': current_users[u['login']]['_id'],
                                     'level': self.access_types[u['type']]
                                     if 'type' in u else u['level']}
                                    for u in users]

        return r.put_access(_id, access_list, public=public)

    def _process_folders(self, folders, parentId, parentType):
        """
        Process a list of folders from a user or collection.

        :param folders: List of folders passed as attribute
                        to user or collection
        :param parentId: ID of the user or the collection
        :param parentType: one of 'user' or 'collection'
        :returns: Nothing
        :rtype: None
        """
        current_folders = {f['name']: f for f in
                           self.get('folder', {'parentType': parentType,
                                               'parentId': parentId})}
        # Add, update or noop listed folders
        for folder in folders:
            # some validation of folder here would be a good idea
            kwargs = folder.copy()
            del kwargs['name']
            self.folder(folder['name'],
                        parentId=parentId,
                        parentType=parentType,
                        **kwargs)

        # Make sure we remove folders not listed
        for name in (set(current_folders.keys())
                     - set([f['name'] for f in folders])):

            original_state = self.module.params['state']
            self.module.params['state'] = 'absent'
            self.folder(name,
                        parentId=parentId,
                        parentType=parentType)
            self.module.params['state'] = original_state

    def collection(self, name, description=None,
                   public=True, access=None, folders=None, debug=False):

        ret = {}
        r = CollectionResource(self)
        valid_fields = [('name', name),
                        ('description', description)]

        if self.module.params['state'] == 'present':
            if r.name_exists(name):
                # While we can set public when we create the collection, we
                # cannot update the public/private status of a collection
                # via the PUT /collection/%s endpoint. Currently this is
                # possible through the API by hitting the
                # PUT /collection/%s/access endpoint with public=true and
                # the access dict equal to {}
                if r.resources_by_name[name]['public'] != public:
                    _id = r.resources_by_name[name]['_id']
                    self.changed = True
                    self._access(r, r.get_access(_id), _id, public=public)
                    # invalidate the resource cache - this forces us to pick up
                    # the change in 'public' attribute despite it not being
                    # an attribute we can modify
                    r._resources = None

                ret = r.update_by_name(name, {k: v for k, v in valid_fields
                                              if v is not None})

            else:
                valid_fields.append(('public', public))
                ret = r.create({k: v for k, v in valid_fields
                                if v is not None})
        if folders is not None:
            self._process_folders(folders, ret['_id'], 'collection')

        if access is not None:
            _id = ret['_id']
            ret['access'] = self._access(r, access, _id, public=public)

        elif self.module.params['state'] == 'absent':
            ret = r.delete_by_name(name)

        return ret

    def user(self, login, password, firstName=None,
             lastName=None, email=None, admin=False, folders=None):

        if self.module.params['state'] == 'present':

            # Fail if we don't have firstName, lastName and email
            for var_name, var in [('firstName', firstName),
                                  ('lastName', lastName), ('email', email)]:
                if var is None:
                    self.fail('%s must be set if state '
                              "is 'present'" % var_name)

            try:
                ret = self.authenticate(username=login,
                                        password=password)

                me = self.get('user/me')

                # List of fields that can actually be updated
                updateable = ['firstName', 'lastName', 'email', 'admin']
                passed_in = [firstName, lastName, email, admin]

                # If there is actually an update to be made
                if set([(k, v) for k, v in me.items() if k in updateable]) ^ \
                   set(zip(updateable, passed_in)):

                    self.put('user/%s' % me['_id'],
                             parameters={
                                 'login': login,
                                 'firstName': firstName,
                                 'lastName': lastName,
                                 'password': password,
                                 'email': email,
                                 'admin': 'true' if admin else 'false'})
                    self.changed = True

                ret = me
            # User does not exist (with this login info)
            except AuthenticationError:
                ret = self.post('user', parameters={
                    'login': login,
                    'firstName': firstName,
                    'lastName': lastName,
                    'password': password,
                    'email': email,
                    'admin': 'true' if admin else 'false'
                })
                self.changed = True

            if folders is not None:
                _id = self.get('resource/lookup',
                               {'path': '/user/{}'.format(login)})['_id']
                self._process_folders(folders, _id, 'user')

        elif self.module.params['state'] == 'absent':
            ret = []
            try:
                ret = self.authenticate(username=login,
                                        password=password)

                me = self.get('user/me')

                self.delete('user/%s' % me['_id'])
                self.changed = True
            # User does not exist (with this login info)
            except AuthenticationError:
                ret = []

        return ret

    # Handles patch correctly by dumping the data as a string before passing
    # it on to requests See:
    # http://docs.python-requests.org/en/master/user/quickstart/#more-complicated-post-requests
    def patch(self, path, parameters=None, data=None):
        super(GirderClientModule, self).patch(path, parameters=parameters,
                                              data=json.dumps(data))

    assetstore_types = {
        'filesystem': 0,
        'gridfs': 1,
        's3': 2,
        'hdfs': 'hdfs',
        'database': 'database'
    }

    def __validate_hdfs_assetstore(self, *args, **kwargs):
        # Check if hdfs plugin is available,  enable it if it isn't
        pass

    def __validate_database_assetstore(self, *args, **kwargs):
        pass

    def assetstore(self, name, type, root=None, db=None, mongohost=None,
                   replicaset='', bucket=None, prefix='', accessKeyId=None,
                   secret=None, service='', host=None,
                   port=None, path=None, user=None, webHdfsPort=None,
                   dbtype=None, dburi=None, readOnly=False, current=False,
                   region=DEFAULT_REGION, inferCredentials=False):

        # Fail if somehow we have an asset type not in assetstore_types
        if type not in self.assetstore_types.keys():
            self.fail('assetstore type %s is not implemented!' % type)

        argument_hash = {
            'filesystem': {'name': name,
                           'type': self.assetstore_types[type],
                           'root': root},
            'gridfs': {'name': name,
                       'type': self.assetstore_types[type],
                       'db': db,
                       'mongohost': mongohost,
                       'replicaset': replicaset},
            's3': {'name': name,
                   'type': self.assetstore_types[type],
                   'bucket': bucket},
            'hdfs': {'name': name,
                     'type': self.assetstore_types[type],
                     'host': host,
                     'port': port,
                     'path': path,
                     'user': user,
                     'webHdfsPort': webHdfsPort},
            'database': {'name': name,
                         'type': self.assetstore_types[type],
                         'dbtype': dbtype,
                         'dburi': dburi}
        }

        # Fail if we don't have all the required attributes
        # for this asset type
        for k, v in argument_hash[type].items():
            if v is None:
                self.fail('assetstores of type '
                          '%s require attribute %s' % (type, k))

        # Set optional arguments in the hash
        argument_hash[type]['readOnly'] = readOnly
        argument_hash[type]['current'] = current
        argument_hash[type]['prefix'] = prefix
        argument_hash[type]['accessKeyId'] = accessKeyId
        argument_hash[type]['secret'] = secret
        argument_hash[type]['service'] = service
        argument_hash[type]['region'] = region
        argument_hash[type]['inferCredentials'] = inferCredentials

        ret = []
        # Get the current assetstores
        assetstores = {a['name']: a for a in self.get('assetstore')}

        self.message['debug']['assetstores'] = assetstores

        # If we want the assetstore to be present
        if self.module.params['state'] == 'present':

            # And the asset store exists
            if name in assetstores.keys():

                id = assetstores[name]['_id']

                ####
                # Fields that could potentially be updated
                #
                # This is necessary because there are fields in the assetstores
                # that do not hash (e.g., capacity) and fields in the
                # argument_hash that are not returned by 'GET' assetstore (e.g.
                # readOnly). We could be more precise about this
                # (e.g., by only checking items that are relevant to this type)
                # but readability suffers.
                updateable = ['root', 'mongohost', 'replicaset', 'bucket',
                              'prefix', 'db', 'accessKeyId', 'secret',
                              'service', 'host', 'port', 'path', 'user',
                              'webHdfsPort', 'current', 'dbtype', 'dburi',
                              'region', 'inferCredentials']

                # tuples of (key,  value) for fields that can be updated
                # in the assetstore
                assetstore_items = set((k, assetstores[name][k])
                                       for k in updateable
                                       if k in assetstores[name].keys())

                # tuples of (key,  value) for fields that can be updated
                # in the argument_hash for this assetstore type
                arg_hash_items = set((k, argument_hash[type][k])
                                     for k in updateable
                                     if k in argument_hash[type].keys())

                # if arg_hash_items not a subset of assetstore_items
                if not arg_hash_items <= assetstore_items:
                    # Update
                    ret = self.put('assetstore/%s' % id,
                                   parameters=argument_hash[type])

                    self.changed = True

            # And the asset store does not exist
            else:
                try:
                    # If __validate_[type]_assetstore exists then call the
                    # function with argument_hash. E.g.,  to check if the
                    # HDFS plugin is enabled
                    getattr(self, '__validate_%s_assetstore' % type
                            )(**argument_hash)
                except AttributeError:
                    pass

                ret = self.post('assetstore',
                                parameters=argument_hash[type])
                self.changed = True
        # If we want the assetstore to be gone
        elif self.module.params['state'] == 'absent':
            # And the assetstore exists
            if name in assetstores.keys():
                id = assetstores[name]['_id']
                ret = self.delete('assetstore/%s' % id,
                                  parameters=argument_hash[type])
                self.changed = True

        return ret

    def setting(self, key, value=None):
        ret = {}

        if value is None:
            value = ''

        if self.module.params['state'] == 'present':
            # Get existing setting value to determine self.changed
            existing_value = self.get('system/setting', parameters={'key': key})

            if existing_value is None:
                existing_value = ''

            params = {
                'key': key,
                'value': json.dumps(value)
            }

            try:
                response = self.put('system/setting', data=params)
            except requests.HTTPError as e:
                self.fail(e.response.json()['message'])

            if response:
                self.changed = unjsonify(existing_value) != unjsonify(value)

            if self.changed:
                ret['previous_value'] = existing_value
                ret['current_value'] = value
            else:
                ret['previous_value'] = ret['current_value'] = existing_value

        else:
            self.fail("%s state isn't supported by Girder setting module." %
                      self.module.params['state'])

        return ret


def main():
    """
    Entry point for ansible girder client module

    :returns: Nothing
    :rtype: NoneType
    """
    # Default spec for initalizing and authenticating
    argument_spec = {
        # __init__
        'host': dict(),
        'port': dict(),
        'apiRoot': dict(),
        'apiUrl': dict(),
        'scheme': dict(),
        'dryrun': dict(),
        'blacklist': dict(),

        # authenticate
        'username': dict(),
        'password': dict(
            no_log=True
        ),
        'token': dict(
            no_log=True
        ),
        'apiKey': dict(
            no_log=True
        ),

        # General
        'state': dict(default='present', choices=['present', 'absent'])
    }

    gcm = GirderClientModule()

    for method in gcm.required_one_of:
        argument_spec[method] = dict(type=gcm.spec[method]['type'])

    module = AnsibleModule(
        argument_spec=argument_spec,
        required_one_of=[gcm.required_one_of,
                         ['token', 'username', 'user', 'apiKey']],
        required_together=[['username', 'password']],
        mutually_exclusive=gcm.required_one_of,
        supports_check_mode=False)

    if not HAS_GIRDER_CLIENT:
        module.fail_json(msg='Could not import GirderClient!')

    try:
        gcm(module)

    except requests.HTTPError as e:
        import traceback
        module.fail_json(msg='%s:%s\n%s\n%s' % (e.__class__, str(e),
                                                e.response.text,
                                                traceback.format_exc()))
    except Exception as e:
        import traceback
        # exc_type, exc_obj, exec_tb = sys.exc_info()
        module.fail_json(msg='%s: %s\n\n%s' % (e.__class__, str(e),
                                               traceback.format_exc()))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from ..describe import Description, autoDescribeRoute
from ..rest import Resource, filtermodel, setResponseHeader, setContentDisposition
from girder.api import access
from girder.constants import AccessType, TokenScope
from girder.exceptions import RestException
from girder.models.folder import Folder as FolderModel
from girder.utility import ziputil
from girder.utility.model_importer import ModelImporter
from girder.utility.progress import ProgressContext


class Folder(Resource):
    """API Endpoint for folders."""

    def __init__(self):
        super(Folder, self).__init__()
        self.resourceName = 'folder'
        self._model = FolderModel()
        self.route('DELETE', (':id',), self.deleteFolder)
        self.route('DELETE', (':id', 'contents'), self.deleteContents)
        self.route('GET', (), self.find)
        self.route('GET', (':id',), self.getFolder)
        self.route('GET', (':id', 'details'), self.getFolderDetails)
        self.route('GET', (':id', 'access'), self.getFolderAccess)
        self.route('GET', (':id', 'download'), self.downloadFolder)
        self.route('GET', (':id', 'rootpath'), self.rootpath)
        self.route('POST', (), self.createFolder)
        self.route('PUT', (':id',), self.updateFolder)
        self.route('PUT', (':id', 'access'), self.updateFolderAccess)
        self.route('POST', (':id', 'copy'), self.copyFolder)
        self.route('PUT', (':id', 'metadata'), self.setMetadata)
        self.route('DELETE', (':id', 'metadata'), self.deleteMetadata)

    @access.public(scope=TokenScope.DATA_READ)
    @filtermodel(model=FolderModel)
    @autoDescribeRoute(
        Description('Search for folders by certain properties.')
        .notes('You must pass either a "folderId" or "text" field '
               'to specify how you are searching for folders.  '
               'If you omit one of these parameters the request will fail and respond : '
               '"Invalid search mode."')
        .responseClass('Folder', array=True)
        .param('parentType', "Type of the folder's parent", required=False,
               enum=['folder', 'user', 'collection'])
        .param('parentId', "The ID of the folder's parent.", required=False)
        .param('text', 'Pass to perform a text search.', required=False)
        .param('name', 'Pass to lookup a folder by exact name match. Must '
               'pass parentType and parentId as well when using this.', required=False)
        .pagingParams(defaultSort='lowerName')
        .errorResponse()
        .errorResponse('Read access was denied on the parent resource.', 403)
    )
    def find(self, parentType, parentId, text, name, limit, offset, sort):
        """
        Get a list of folders with given search parameters. Currently accepted
        search modes are:

        1. Searching by parentId and parentType, with optional additional
           filtering by the name field (exact match) or using full text search
           within a single parent folder. Pass a "name" parameter or "text"
           parameter to invoke these additional filters.
        2. Searching with full text search across all folders in the system.
           Simply pass a "text" parameter for this mode.
        """
        user = self.getCurrentUser()

        if parentType and parentId:
            parent = ModelImporter.model(parentType).load(
                parentId, user=user, level=AccessType.READ, exc=True)

            filters = {}
            if text:
                filters['$text'] = {
                    '$search': text
                }
            if name:
                filters['name'] = name

            return self._model.childFolders(
                parentType=parentType, parent=parent, user=user,
                offset=offset, limit=limit, sort=sort, filters=filters)
        elif text:
            return self._model.textSearch(
                text, user=user, limit=limit, offset=offset, sort=sort)
        else:
            raise RestException('Invalid search mode.')

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Get detailed information about a folder.')
        .modelParam('id', model=FolderModel, level=AccessType.READ)
        .errorResponse()
        .errorResponse('Read access was denied on the folder.', 403)
    )
    def getFolderDetails(self, folder):
        return {
            'nItems': self._model.countItems(folder),
            'nFolders': self._model.countFolders(
                folder, user=self.getCurrentUser(), level=AccessType.READ)
        }

    @access.public(scope=TokenScope.DATA_READ, cookie=True)
    @autoDescribeRoute(
        Description('Download an entire folder as a zip archive.')
        .modelParam('id', model=FolderModel, level=AccessType.READ)
        .jsonParam('mimeFilter', 'JSON list of MIME types to include.', required=False,
                   requireArray=True)
        .produces('application/zip')
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the folder.', 403)
    )
    def downloadFolder(self, folder, mimeFilter):
        """
        Returns a generator function that will be used to stream out a zip
        file containing this folder's contents, filtered by permissions.
        """
        setResponseHeader('Content-Type', 'application/zip')
        setContentDisposition(folder['name'] + '.zip')
        user = self.getCurrentUser()

        def stream():
            zip = ziputil.ZipGenerator(folder['name'])
            for (path, file) in self._model.fileList(
                    folder, user=user, subpath=False, mimeFilter=mimeFilter):
                for data in zip.addFile(file, path):
                    yield data
            yield zip.footer()
        return stream

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model=FolderModel)
    @autoDescribeRoute(
        Description('Update a folder or move it into a new parent.')
        .responseClass('Folder')
        .modelParam('id', model=FolderModel, level=AccessType.WRITE)
        .param('name', 'Name of the folder.', required=False, strip=True)
        .param('description', 'Description for the folder.', required=False, strip=True)
        .param('parentType', "Type of the folder's parent", required=False,
               enum=['folder', 'user', 'collection'], strip=True)
        .param('parentId', 'Parent ID for the new parent of this folder.', required=False)
        .jsonParam('metadata', 'A JSON object containing the metadata keys to add',
                   paramType='form', requireObject=True, required=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Write access was denied for the folder or its new parent object.', 403)
    )
    def updateFolder(self, folder, name, description, parentType, parentId, metadata):
        user = self.getCurrentUser()
        if name is not None:
            folder['name'] = name
        if description is not None:
            folder['description'] = description

        folder = self._model.updateFolder(folder)
        if metadata:
            folder = self._model.setMetadata(folder, metadata)

        if parentType and parentId:
            parent = ModelImporter.model(parentType).load(
                parentId, level=AccessType.WRITE, user=user, exc=True)
            if (parentType, parent['_id']) != (folder['parentCollection'], folder['parentId']):
                folder = self._model.move(folder, parent, parentType)

        return folder

    @access.user(scope=TokenScope.DATA_OWN)
    @filtermodel(model=FolderModel, addFields={'access'})
    @autoDescribeRoute(
        Description('Update the access control list for a folder.')
        .modelParam('id', model=FolderModel, level=AccessType.ADMIN)
        .jsonParam('access', 'The JSON-encoded access control list.', requireObject=True)
        .jsonParam('publicFlags', 'JSON list of public access flags.', requireArray=True,
                   required=False)
        .param('public', 'Whether the folder should be publicly visible.',
               dataType='boolean', required=False)
        .param('recurse', 'Whether the policies should be applied to all '
               'subfolders under this folder as well.', dataType='boolean',
               default=False, required=False)
        .param('progress', 'If recurse is set to True, this controls whether '
               'progress notifications will be sent.', dataType='boolean',
               default=False, required=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin access was denied for the folder.', 403)
    )
    def updateFolderAccess(self, folder, access, publicFlags, public, recurse, progress):
        user = self.getCurrentUser()
        progress = progress and recurse  # Only enable progress in recursive case
        with ProgressContext(progress, user=user, title='Updating permissions',
                             message='Calculating progress...') as ctx:
            if progress:
                ctx.update(total=self._model.subtreeCount(
                    folder, includeItems=False, user=user, level=AccessType.ADMIN))
            return self._model.setAccessList(
                folder, access, save=True, recurse=recurse, user=user,
                progress=ctx, setPublic=public, publicFlags=publicFlags)

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model=FolderModel)
    @autoDescribeRoute(
        Description('Create a new folder.')
        .responseClass('Folder')
        .param('parentType', "Type of the folder's parent", required=False,
               enum=['folder', 'user', 'collection'], default='folder')
        .param('parentId', "The ID of the folder's parent.")
        .param('name', 'Name of the folder.', strip=True)
        .param('description', 'Description for the folder.', required=False,
               default='', strip=True)
        .param('reuseExisting', 'Return existing folder if it exists rather than '
               'creating a new one.', required=False,
               dataType='boolean', default=False)
        .param('public', 'Whether the folder should be publicly visible. By '
               'default, inherits the value from parent folder, or in the '
               'case of user or collection parentType, defaults to False.',
               required=False, dataType='boolean')
        .jsonParam('metadata', 'A JSON object containing the metadata keys to add',
                   paramType='form', requireObject=True, required=False)
        .errorResponse()
        .errorResponse('Write access was denied on the parent', 403)
    )
    def createFolder(self, public, parentType, parentId, name, description,
                     reuseExisting, metadata):
        user = self.getCurrentUser()
        parent = ModelImporter.model(parentType).load(
            id=parentId, user=user, level=AccessType.WRITE, exc=True)

        newFolder = self._model.createFolder(
            parent=parent, name=name, parentType=parentType, creator=user,
            description=description, public=public, reuseExisting=reuseExisting)
        if metadata:
            newFolder = self._model.setMetadata(newFolder, metadata)
        return newFolder

    @access.public(scope=TokenScope.DATA_READ)
    @filtermodel(model=FolderModel)
    @autoDescribeRoute(
        Description('Get a folder by ID.')
        .responseClass('Folder')
        .modelParam('id', model=FolderModel, level=AccessType.READ)
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the folder.', 403)
    )
    def getFolder(self, folder):
        return folder

    @access.user(scope=TokenScope.DATA_OWN)
    @autoDescribeRoute(
        Description('Get the access control list for a folder.')
        .responseClass('Folder')
        .modelParam('id', model=FolderModel, level=AccessType.ADMIN)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin access was denied for the folder.', 403)
    )
    def getFolderAccess(self, folder):
        return self._model.getFullAccessList(folder)

    @access.user(scope=TokenScope.DATA_OWN)
    @autoDescribeRoute(
        Description('Delete a folder by ID.')
        .modelParam('id', model=FolderModel, level=AccessType.ADMIN)
        .param('progress', 'Whether to record progress on this task.',
               required=False, dataType='boolean', default=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin access was denied for the folder.', 403)
    )
    def deleteFolder(self, folder, progress):
        with ProgressContext(progress, user=self.getCurrentUser(),
                             title='Deleting folder %s' % folder['name'],
                             message='Calculating folder size...') as ctx:
            # Don't do the subtree count if we weren't asked for progress
            if progress:
                ctx.update(total=self._model.subtreeCount(folder))
            self._model.remove(folder, progress=ctx)
        return {'message': 'Deleted folder %s.' % folder['name']}

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model=FolderModel)
    @autoDescribeRoute(
        Description('Set metadata fields on an folder.')
        .responseClass('Folder')
        .notes('Set metadata fields to null in order to delete them.')
        .modelParam('id', model=FolderModel, level=AccessType.WRITE)
        .jsonParam('metadata', 'A JSON object containing the metadata keys to add',
                   paramType='body', requireObject=True)
        .param('allowNull', 'Whether "null" is allowed as a metadata value.', required=False,
               dataType='boolean', default=False)
        .errorResponse(('ID was invalid.',
                        'Invalid JSON passed in request body.',
                        'Metadata key name was invalid.'))
        .errorResponse('Write access was denied for the folder.', 403)
    )
    def setMetadata(self, folder, metadata, allowNull):
        return self._model.setMetadata(folder, metadata, allowNull=allowNull)

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model=FolderModel)
    @autoDescribeRoute(
        Description('Copy a folder.')
        .responseClass('Folder')
        .modelParam('id', 'The ID of the original folder.', model=FolderModel,
                    level=AccessType.READ)
        .param('parentType', "Type of the new folder's parent", required=False,
               enum=['folder', 'user', 'collection'])
        .param('parentId', 'The ID of the parent document.', required=False)
        .param('name', 'Name for the new folder.', required=False)
        .param('description', 'Description for the new folder.', required=False)
        .param('public', 'Whether the folder should be publicly visible. By '
               'default, inherits the value from parent folder, or in the case '
               'of user or collection parentType, defaults to False. If '
               "'original', use the value of the original folder.",
               required=False, enum=['true', 'false', 'original'])
        .param('progress', 'Whether to record progress on this task.',
               required=False, dataType='boolean', default=False)
        .errorResponse(('A parameter was invalid.',
                        'ID was invalid.'))
        .errorResponse('Read access was denied on the original folder.\n\n'
                       'Write access was denied on the parent.', 403)
    )
    def copyFolder(self, folder, parentType, parentId, name, description, public, progress):
        user = self.getCurrentUser()
        parentType = parentType or folder['parentCollection']
        if parentId:
            parent = ModelImporter.model(parentType).load(
                id=parentId, user=user, level=AccessType.WRITE, exc=True)
        else:
            parent = None

        with ProgressContext(progress, user=self.getCurrentUser(),
                             title='Copying folder %s' % folder['name'],
                             message='Calculating folder size...') as ctx:
            # Don't do the subtree count if we weren't asked for progress
            if progress:
                ctx.update(total=self._model.subtreeCount(folder))
            return self._model.copyFolder(
                folder, creator=user, name=name, parentType=parentType,
                parent=parent, description=description, public=public, progress=ctx)

    @access.user(scope=TokenScope.DATA_WRITE)
    @autoDescribeRoute(
        Description('Remove all contents from a folder.')
        .notes('Cleans out all the items and subfolders from under a folder, '
               'but does not remove the folder itself.')
        .modelParam('id', 'The ID of the folder to clean.', model=FolderModel,
                    level=AccessType.WRITE)
        .param('progress', 'Whether to record progress on this task.',
               required=False, dataType='boolean', default=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Write access was denied on the folder.', 403)
    )
    def deleteContents(self, folder, progress):
        with ProgressContext(progress, user=self.getCurrentUser(),
                             title='Clearing folder %s' % folder['name'],
                             message='Calculating folder size...') as ctx:
            # Don't do the subtree count if we weren't asked for progress
            if progress:
                ctx.update(total=self._model.subtreeCount(folder) - 1)
            self._model.clean(folder, progress=ctx)
        return {'message': 'Cleaned folder %s.' % folder['name']}

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(FolderModel)
    @autoDescribeRoute(
        Description('Delete metadata fields on a folder.')
        .responseClass('Folder')
        .modelParam('id', model=FolderModel, level=AccessType.WRITE)
        .jsonParam(
            'fields', 'A JSON list containing the metadata fields to delete',
            paramType='body', schema={
                'type': 'array',
                'items': {
                    'type': 'string'
                }
            }
        )
        .errorResponse(('ID was invalid.',
                        'Invalid JSON passed in request body.',
                        'Metadata key name was invalid.'))
        .errorResponse('Write access was denied for the folder.', 403)
    )
    def deleteMetadata(self, folder, fields):
        return self._model.deleteMetadata(folder, fields)

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description("Get the path to the root of the folder's hierarchy.")
        .modelParam('id', model=FolderModel, level=AccessType.READ)
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the folder.', 403)
    )
    def rootpath(self, folder, params):
        return self._model.parentsToRoot(folder, user=self.getCurrentUser())



from girder.exceptions import ValidationException
from girder.utility import setting_utilities


class PluginSettings(object):
    DEFAULT_COLLECTION_QUOTA = 'user_quota.default_collection_quota'
    DEFAULT_USER_QUOTA = 'user_quota.default_user_quota'


@setting_utilities.default((
    PluginSettings.DEFAULT_COLLECTION_QUOTA,
    PluginSettings.DEFAULT_USER_QUOTA
))
def _defaultSettings():
    return None


@setting_utilities.validator((
    PluginSettings.DEFAULT_COLLECTION_QUOTA,
    PluginSettings.DEFAULT_USER_QUOTA
))
def _validateSettings(doc):
    from .quota import ValidateSizeQuota
    val = doc['value']

    val, err = ValidateSizeQuota(val)
    if err:
        raise ValidationException(err, 'value')
    doc['value'] = val



# -*- coding: utf-8 -*-
import os

from setuptools import find_packages, setup


def prerelease_local_scheme(version):
    """Return local scheme version unless building on master in CircleCI.
    This function returns the local scheme version number
    (e.g. 0.0.0.dev<N>+g<HASH>) unless building on CircleCI for a
    pre-release in which case it ignores the hash and produces a
    PEP440 compliant pre-release version number (e.g. 0.0.0.dev<N>).
    """
    from setuptools_scm.version import get_local_node_and_date

    if os.getenv('CIRCLE_BRANCH') == 'master':
        return ''
    else:
        return get_local_node_and_date(version)


setup(
    name='girder-audit-logs',
    use_scm_version={'root': '../..', 'local_scheme': prerelease_local_scheme},
    setup_requires=['setuptools-scm'],
    description='Keeps detailed logs of every REST request and low-level file download event.',
    author='Kitware, Inc.',
    author_email='kitware@kitware.com',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6'
    ],
    packages=find_packages(),
    zip_safe=False,
    install_requires=['girder>=3'],
    entry_points={
        'girder.plugin': [
            'audit_logs = girder_audit_logs:AuditLogsPlugin'
        ],
        'girder.cli_plugins': [
            'audit-logs-cleanup = girder_audit_logs.cleanup:cleanup',
            'audit-logs-report = girder_audit_logs.report:report'
        ]
    }
)



# -*- coding: utf-8 -*-
import hashlib
import six

import girder
from girder import events
from girder.api import access
from girder.api.describe import autoDescribeRoute, Description
from girder.api.rest import (
    filtermodel, setRawResponse, setResponseHeader, setContentDisposition)
from girder.api.v1.file import File
from girder.constants import AccessType, TokenScope
from girder.exceptions import RestException
from girder.models.file import File as FileModel
from girder.models.setting import Setting
from girder.plugin import GirderPlugin
from girder.utility.progress import ProgressContext, noProgress

from .settings import PluginSettings


SUPPORTED_ALGORITHMS = {'sha512'}
_CHUNK_LEN = 65536


class HashedFile(File):
    @property
    def supportedAlgorithms(self):
        girder.logger.warning(
            'HashedFile.supportedAlgorithms is deprecated, use the module-level '
            'SUPPORTED_ALGORITHMS instead.')
        return SUPPORTED_ALGORITHMS

    def __init__(self, node):
        super(HashedFile, self).__init__()

        node.route('GET', ('hashsum', ':algo', ':hash'), self.getByHash)
        node.route('GET', ('hashsum', ':algo', ':hash', 'download'), self.downloadWithHash)
        node.route('GET', (':id', 'hashsum_file', ':algo'), self.downloadKeyFile)
        node.route('POST', (':id', 'hashsum'), self.computeHashes)

    @access.public(scope=TokenScope.DATA_READ, cookie=True)
    @autoDescribeRoute(
        Description('Download the hashsum key file for a given file.')
        .modelParam('id', 'The ID of the file.', model=FileModel, level=AccessType.READ)
        .param('algo', 'The hashsum algorithm.', paramType='path', lower=True,
               enum=SUPPORTED_ALGORITHMS)
        .notes("This is meant to be used in conjunction with CMake's ExternalData module.")
        .produces('text/plain')
        .errorResponse()
        .errorResponse('Read access was denied on the file.', 403)
    )
    def downloadKeyFile(self, file, algo):
        self._validateAlgo(algo)

        if algo not in file:
            raise RestException('This file does not have the %s hash computed.' % algo)
        keyFileBody = '%s\n' % file[algo]
        name = '.'.join((file['name'], algo))

        setResponseHeader('Content-Length', len(keyFileBody))
        setResponseHeader('Content-Type', 'text/plain')
        setContentDisposition(name)
        setRawResponse()

        return keyFileBody

    @access.public(scope=TokenScope.DATA_READ, cookie=True)
    @autoDescribeRoute(
        Description('Download a file by its hashsum.')
        .param('algo', 'The type of the given hashsum (case insensitive).',
               paramType='path', lower=True, enum=SUPPORTED_ALGORITHMS)
        .param('hash', 'The hexadecimal hashsum of the file to download (case insensitive).',
               paramType='path', lower=True)
        .errorResponse('No file with the given hash exists.')
    )
    def downloadWithHash(self, algo, hash, params):
        file = self._getFirstFileByHash(algo, hash)
        if not file:
            raise RestException('File not found.', code=404)

        return self.download(id=file['_id'], params=params)

    @access.public(scope=TokenScope.DATA_READ, cookie=True)
    @filtermodel(FileModel)
    @autoDescribeRoute(
        Description('Return a list of files matching a hashsum.')
        .param('algo', 'The type of the given hashsum (case insensitive).',
               paramType='path', lower=True, enum=SUPPORTED_ALGORITHMS)
        .param('hash', 'The hexadecimal hashsum of the file to download (case insensitive).',
               paramType='path', lower=True)
    )
    def getByHash(self, algo, hash):
        self._validateAlgo(algo)

        model = FileModel()
        user = self.getCurrentUser()
        cursor = model.find({algo: hash})
        return [file for file in cursor if model.hasAccess(file, user, AccessType.READ)]

    @access.user(scope=TokenScope.DATA_WRITE)
    @autoDescribeRoute(
        Description('Manually compute the checksum values for a given file.')
        .modelParam('id', 'The ID of the file.', model=FileModel, level=AccessType.WRITE)
        .param('progress', 'Whether to track progress of the operation', dataType='boolean',
               default=False, required=False)
        .errorResponse()
        .errorResponse('Write access was denied on the file.', 403)
    )
    def computeHashes(self, file, progress):
        with ProgressContext(
                progress, title='Computing hash: %s' % file['name'], total=file['size'],
                user=self.getCurrentUser()) as pc:
            return _computeHash(file, progress=pc)

    def _validateAlgo(self, algo):
        """
        Print an exception if a user requests an invalid checksum algorithm.
        """
        if algo not in SUPPORTED_ALGORITHMS:
            msg = 'Invalid algorithm "%s". Supported algorithms: %s.' % (
                algo, ', '.join(SUPPORTED_ALGORITHMS))
            raise RestException(msg, code=400)

    def _getFirstFileByHash(self, algo, hash, user=None):
        """
        Return the first file that the user has access to given its hash and its
        associated hashsum algorithm name.

        :param algo: Algorithm the given hash is encoded with.
        :param hash: Hash of the file to find.
        :param user: User to test access against.
         Default (none) is the current user.
        :return: A file document.
        """
        self._validateAlgo(algo)

        query = {algo: hash}
        fileModel = FileModel()
        cursor = fileModel.find(query)

        if not user:
            user = self.getCurrentUser()

        for file in cursor:
            if fileModel.hasAccess(file, user, AccessType.READ):
                return file

        return None


def _computeHashHook(event):
    """
    Event hook that computes the file hashes in the background after
    a completed upload. Only done if the AUTO_COMPUTE setting enabled.
    """
    if Setting().get(PluginSettings.AUTO_COMPUTE):
        _computeHash(event.info['file'])


def _computeHash(file, progress=noProgress):
    """
    Computes all supported checksums on a given file. Downloads the
    file data and stream-computes all required hashes on it, saving
    the results in the file document.

    In the case of assetstore impls that already compute the sha512,
    and when sha512 is the only supported algorithm, we will not download
    the file to the server.
    """
    toCompute = SUPPORTED_ALGORITHMS - set(file)
    toCompute = {alg: getattr(hashlib, alg)() for alg in toCompute}

    if not toCompute:
        return

    fileModel = FileModel()
    with fileModel.open(file) as fh:
        while True:
            chunk = fh.read(_CHUNK_LEN)
            if not chunk:
                break
            for digest in six.viewvalues(toCompute):
                digest.update(chunk)
            progress.update(increment=len(chunk))

    digests = {alg: digest.hexdigest() for alg, digest in six.viewitems(toCompute)}
    fileModel.update({'_id': file['_id']}, update={
        '$set': digests
    }, multi=False)

    return digests


class HashsumDownloadPlugin(GirderPlugin):
    DISPLAY_NAME = 'Hashsum download'
    CLIENT_SOURCE_PATH = 'web_client'

    def load(self, info):
        HashedFile(info['apiRoot'].file)
        FileModel().exposeFields(level=AccessType.READ, fields=SUPPORTED_ALGORITHMS)

        events.bind('data.process', 'hashsum_download', _computeHashHook)



import os
from .fixtures import *  # noqa
from .fixtures import _disableRealDatabaseConnectivity  # noqa


def _makeCoverageDirs(config):
    """
    Create the necessary directories for coverage. This is necessary because neither coverage nor
    pytest-cov have support for making the data_file directory before running.
    """
    covPlugin = config.pluginmanager.get_plugin('_cov')

    if covPlugin is not None:
        covPluginConfig = covPlugin.cov_controller.cov.config
        covDataFileDir = os.path.dirname(covPluginConfig.data_file)

        try:
            os.makedirs(covDataFileDir)
        except OSError:
            pass


def _addCustomMarkers(config):
    markerDocs = [
        'plugin(pluginName, [pluginClass]): load a plugin (may be marked multiple times)',
    ]
    for markerDoc in markerDocs:
        config.addinivalue_line('markers', markerDoc)


def pytest_configure(config):
    _makeCoverageDirs(config)
    _addCustomMarkers(config)


def pytest_addoption(parser):
    group = parser.getgroup('girder')
    group.addoption('--mock-db', action='store_true', default=False,
                    help='Whether or not to mock the database using mongomock.')
    group.addoption('--mongo-uri', action='store', default='mongodb://localhost:27017',
                    help=('The base URI to the MongoDB instance to use for database connections, '
                          'default is mongodb://localhost:27017'))
    group.addoption('--keep-db', action='store_true', default=False,
                    help='Whether to destroy testing databases after running tests.')



# -*- coding: utf-8 -*-
import cherrypy
import girder
from girder.utility import server

cherrypy.config.update({'engine.autoreload.on': False,
                        'environment': 'embedded'})
cherrypy.config['server'].update({'disable_event_daemon': True})

# TODO The below line can be removed if we do away with girder.logprint
girder._quiet = True  # This means we won't duplicate messages to stdout/stderr
_formatter = girder.LogFormatter('[%(asctime)s] %(levelname)s: %(message)s')
_handler = cherrypy._cplogging.WSGIErrorHandler()
_handler.setFormatter(_formatter)
girder.logger.addHandler(_handler)

# 'application' is the default callable object for WSGI implementations, see PEP 3333 for more.
server.setup()
application = cherrypy.tree

cherrypy.server.unsubscribe()
cherrypy.engine.start()



# -*- coding: utf-8 -*-
from ..describe import Description, autoDescribeRoute
from ..rest import Resource, filtermodel, setResponseHeader, setContentDisposition
from girder.api import access
from girder.constants import AccessType, TokenScope
from girder.models.collection import Collection as CollectionModel
from girder.exceptions import AccessException
from girder.utility import ziputil
from girder.utility.progress import ProgressContext


class Collection(Resource):
    """API Endpoint for collections."""

    def __init__(self):
        super(Collection, self).__init__()
        self.resourceName = 'collection'
        self._model = CollectionModel()

        self.route('DELETE', (':id',), self.deleteCollection)
        self.route('GET', (), self.find)
        self.route('GET', (':id',), self.getCollection)
        self.route('GET', (':id', 'details'), self.getCollectionDetails)
        self.route('GET', ('details',), self.getCollectionsDetails)
        self.route('GET', (':id', 'download'), self.downloadCollection)
        self.route('GET', (':id', 'access'), self.getCollectionAccess)
        self.route('POST', (), self.createCollection)
        self.route('PUT', (':id',), self.updateCollection)
        self.route('PUT', (':id', 'access'), self.updateCollectionAccess)
        self.route('PUT', (':id', 'metadata'), self.setMetadata)
        self.route('DELETE', (':id', 'metadata'), self.deleteMetadata)

    @access.public(scope=TokenScope.DATA_READ)
    @filtermodel(model=CollectionModel)
    @autoDescribeRoute(
        Description('List or search for collections.')
        .responseClass('Collection', array=True)
        .param('text', 'Pass this to perform a text search for collections.', required=False)
        .pagingParams(defaultSort='name')
    )
    def find(self, text, limit, offset, sort):
        user = self.getCurrentUser()

        if text is not None:
            return self._model.textSearch(text, user=user, limit=limit, offset=offset)

        return self._model.list(user=user, offset=offset, limit=limit, sort=sort)

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model=CollectionModel)
    @autoDescribeRoute(
        Description('Create a new collection.')
        .responseClass('Collection')
        .param('name', 'Name for the collection. Must be unique.')
        .param('description', 'Collection description.', required=False)
        .param('public', 'Whether the collection should be publicly visible.',
               required=False, dataType='boolean', default=False)
        .errorResponse()
        .errorResponse('You are not authorized to create collections.', 403)
    )
    def createCollection(self, name, description, public):
        user = self.getCurrentUser()

        if not self._model.hasCreatePrivilege(user):
            raise AccessException('You are not authorized to create collections.')

        return self._model.createCollection(
            name=name, description=description, public=public, creator=user)

    @access.public(scope=TokenScope.DATA_READ)
    @filtermodel(model=CollectionModel)
    @autoDescribeRoute(
        Description('Get a collection by ID.')
        .responseClass('Collection')
        .modelParam('id', model=CollectionModel, level=AccessType.READ)
        .errorResponse('ID was invalid.')
        .errorResponse('Read permission denied on the collection.', 403)
    )
    def getCollection(self, collection):
        return collection

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Get detailed information of accessible collections.')
    )
    def getCollectionsDetails(self):
        count = self._model.findWithPermissions(user=self.getCurrentUser()).count()
        return {
            'nCollections': count
        }

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Get detailed information about a collection.')
        .modelParam('id', model=CollectionModel, level=AccessType.READ)
        .errorResponse()
        .errorResponse('Read access was denied on the collection.', 403)
    )
    def getCollectionDetails(self, collection):
        return {
            'nFolders': self._model.countFolders(
                collection, user=self.getCurrentUser(), level=AccessType.READ)
        }

    @access.public(scope=TokenScope.DATA_READ, cookie=True)
    @autoDescribeRoute(
        Description('Download an entire collection as a zip archive.')
        .modelParam('id', model=CollectionModel, level=AccessType.READ)
        .jsonParam('mimeFilter', 'JSON list of MIME types to include.', requireArray=True,
                   required=False)
        .produces('application/zip')
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the collection.', 403)
    )
    def downloadCollection(self, collection, mimeFilter):
        setResponseHeader('Content-Type', 'application/zip')
        setContentDisposition(collection['name'] + '.zip')

        def stream():
            zip = ziputil.ZipGenerator(collection['name'])
            for (path, file) in self._model.fileList(
                    collection, user=self.getCurrentUser(), subpath=False, mimeFilter=mimeFilter):
                for data in zip.addFile(file, path):
                    yield data
            yield zip.footer()
        return stream

    @access.user(scope=TokenScope.DATA_OWN)
    @autoDescribeRoute(
        Description('Get the access control list for a collection.')
        .modelParam('id', model=CollectionModel, level=AccessType.ADMIN)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin permission denied on the collection.', 403)
    )
    def getCollectionAccess(self, collection):
        return self._model.getFullAccessList(collection)

    @access.user(scope=TokenScope.DATA_OWN)
    @filtermodel(model=CollectionModel, addFields={'access'})
    @autoDescribeRoute(
        Description('Set the access control list for a collection.')
        .modelParam('id', model=CollectionModel, level=AccessType.ADMIN)
        .jsonParam('access', 'The access control list as JSON.', requireObject=True)
        .jsonParam('publicFlags', 'List of public access flags to set on the collection.',
                   required=False, requireArray=True)
        .param('public', 'Whether the collection should be publicly visible.',
               dataType='boolean', required=False)
        .param('recurse', 'Whether the policies should be applied to all '
               'folders under this collection as well.', dataType='boolean',
               default=False, required=False)
        .param('progress', 'If recurse is set to True, this controls whether '
               'progress notifications will be sent.', dataType='boolean',
               default=False, required=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin permission denied on the collection.', 403)
    )
    def updateCollectionAccess(self, collection, access, public, recurse, progress, publicFlags):
        user = self.getCurrentUser()
        progress = progress and recurse

        with ProgressContext(progress, user=user, title='Updating permissions',
                             message='Calculating progress...') as ctx:
            if progress:
                ctx.update(total=self._model.subtreeCount(
                    collection, includeItems=False, user=user, level=AccessType.ADMIN))
            return self._model.setAccessList(
                collection, access, save=True, user=user, recurse=recurse,
                progress=ctx, setPublic=public, publicFlags=publicFlags)

    @access.user(scope=TokenScope.DATA_READ)
    @filtermodel(model=CollectionModel)
    @autoDescribeRoute(
        Description('Edit a collection by ID.')
        .responseClass('Collection')
        .modelParam('id', model=CollectionModel, level=AccessType.WRITE)
        .param('name', 'Unique name for the collection.', required=False, strip=True)
        .param('description', 'Collection description.', required=False, strip=True)
        .errorResponse('ID was invalid.')
        .errorResponse('Write permission denied on the collection.', 403)
    )
    def updateCollection(self, collection, name, description):
        if name is not None:
            collection['name'] = name
        if description is not None:
            collection['description'] = description

        return self._model.updateCollection(collection)

    @access.user(scope=TokenScope.DATA_OWN)
    @autoDescribeRoute(
        Description('Delete a collection by ID.')
        .modelParam('id', model=CollectionModel, level=AccessType.ADMIN)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin permission denied on the collection.', 403)
    )
    def deleteCollection(self, collection):
        self._model.remove(collection)
        return {'message': 'Deleted collection %s.' % collection['name']}

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model=CollectionModel)
    @autoDescribeRoute(
        Description('Set metadata fields on a collection.')
        .responseClass('Collection')
        .notes('Set metadata fields to null in order to delete them.')
        .modelParam('id', model=CollectionModel, level=AccessType.WRITE)
        .jsonParam('metadata', 'A JSON object containing the metadata keys to add',
                   paramType='body', requireObject=True)
        .param('allowNull', 'Whether "null" is allowed as a metadata value.', required=False,
               dataType='boolean', default=False)
        .errorResponse(('ID was invalid.',
                        'Invalid JSON passed in request body.',
                        'Metadata key name was invalid.'))
        .errorResponse('Write access was denied for the collection.', 403)
    )
    def setMetadata(self, collection, metadata, allowNull):
        return self._model.setMetadata(collection, metadata, allowNull=allowNull)

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(CollectionModel)
    @autoDescribeRoute(
        Description('Delete metadata fields on a collection.')
        .responseClass('Collection')
        .modelParam('id', model=CollectionModel, level=AccessType.WRITE)
        .jsonParam(
            'fields', 'A JSON list containing the metadata fields to delete',
            paramType='body', schema={
                'type': 'array',
                'items': {
                    'type': 'string'
                }
            }
        )
        .errorResponse(('ID was invalid.',
                        'Invalid JSON passed in request body.',
                        'Metadata key name was invalid.'))
        .errorResponse('Write access was denied for the collection.', 403)
    )
    def deleteMetadata(self, collection, fields):
        return self._model.deleteMetadata(collection, fields)



import pydas

MIDAS_URL = 'http://127.0.0.1'
MIDAS_LOGIN = 'user@example.com'
MIDAS_API_KEY = 'API_KEY'

token = pydas.login(email=MIDAS_LOGIN, api_key=MIDAS_API_KEY, url=MIDAS_URL)
mc = pydas.session.communicator


def handle_item(item, bc):
    bc = bc + (item['name'],)
    print('/'.join(bc))


def handle_folder(folder, bc):
    bc = bc + (folder['name'],)
    children = mc.folder_children(token, folder['folder_id'])
    folders = children['folders']
    items = children['items']
    for child in folders:
        handle_folder(child, bc)
    for item in items:
        handle_item(item, bc)


def handle_community(community):
    bc = ('collection', community['name'])
    children = mc.get_community_children(community['community_id'], token)
    folders = children['folders']
    for folder in folders:
        handle_folder(folder, bc)


def handle_user(user):
    bc = ('user', user['email'])
    user_folder = mc.folder_get(token, user['folder_id'])
    children = mc.folder_children(token, user_folder['folder_id'])
    for folder in children['folders']:
        handle_folder(folder, bc)


def main():
    users = mc.list_users(limit=0)
    for user in users:
        handle_user(user)
    communities = mc.list_communities(token)
    for community in communities:
        handle_community(community)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Constants should be defined here.
"""
import os
import sys

import girder

PACKAGE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.dirname(PACKAGE_DIR)
LOG_ROOT = os.path.join(os.path.expanduser('~'), '.girder', 'logs')
MAX_LOG_SIZE = 1024 * 1024 * 10  # Size in bytes before logs are rotated.
LOG_BACKUP_COUNT = 5
ACCESS_FLAGS = {}

# Identifier for Girder's entry in the route table
GIRDER_ROUTE_ID = 'core_girder'

# Threshold below which text search results will be sorted by their text score.
# Setting this too high causes mongodb to use too many resources for searches
# that yield lots of results.
TEXT_SCORE_SORT_MAX = 200
VERSION = {
    'release': girder.__version__
}

#: The local directory containing the static content.
STATIC_PREFIX = os.path.join(sys.prefix, 'share', 'girder')
STATIC_ROOT_DIR = os.path.join(STATIC_PREFIX, 'static')


def registerAccessFlag(key, name, description=None, admin=False):
    """
    Register a new access flag in the set of ACCESS_FLAGS available
    on data in the hierarchy. These are boolean switches that can be used
    to control access to specific functionality on specific resoruces.

    :param key: The unique identifier for this access flag.
    :type key: str
    :param name: Human readable name for this permission (displayed in UI).
    :type name: str
    :param description: Human readable longer description for the flag.
    :type description: str
    :param admin: Set this to True to only allow site admin users to set
        this flag. If True, the flag will only appear in the list for
        site admins. This can be useful for flags with security
        considerations.
    """
    ACCESS_FLAGS[key] = {
        'name': name,
        'description': description,
        'admin': admin
    }


class ServerMode(object):
    PRODUCTION = 'production'
    DEVELOPMENT = 'development'
    TESTING = 'testing'


class TerminalColor(object):
    """
    Provides a set of values that can be used to color text in the terminal.
    """

    ERROR = '\033[1;91m'
    SUCCESS = '\033[32m'
    WARNING = '\033[1;33m'
    INFO = '\033[35m'
    ENDC = '\033[0m'

    @staticmethod
    def _color(tag, text):
        return ''.join([tag, text, TerminalColor.ENDC])

    @staticmethod
    def error(text):
        return TerminalColor._color(TerminalColor.ERROR, text)

    @staticmethod
    def success(text):
        return TerminalColor._color(TerminalColor.SUCCESS, text)

    @staticmethod
    def warning(text):
        return TerminalColor._color(TerminalColor.WARNING, text)

    @staticmethod
    def info(text):
        return TerminalColor._color(TerminalColor.INFO, text)


class AssetstoreType(object):
    """
    All possible assetstore implementation types.
    """

    FILESYSTEM = 0
    GRIDFS = 1
    S3 = 2


class AccessType(object):
    """
    Represents the level of access granted to a user or group on an
    AccessControlledModel. Having a higher access level on a resource also
    confers all of the privileges of the lower levels.

    Semantically, READ access on a resource means that the user can see all
    the information pertaining to the resource, but cannot modify it.

    WRITE access usually means the user can modify aspects of the resource.

    ADMIN access confers total control; the user can delete the resource and
    also manage permissions for other users on it.
    """

    NONE = -1
    READ = 0
    WRITE = 1
    ADMIN = 2
    SITE_ADMIN = 100

    @classmethod
    def validate(cls, level):
        level = int(level)
        if level in (cls.NONE, cls.READ, cls.WRITE, cls.ADMIN, cls.SITE_ADMIN):
            return level
        else:
            raise ValueError('Invalid AccessType: %d.' % level)


class SortDir(object):
    ASCENDING = 1
    DESCENDING = -1


class TokenScope(object):
    """
    Constants for core token scope strings. Token scopes must not contain
    spaces, since many services accept scope lists as a space-separated list
    of strings.
    """

    ANONYMOUS_SESSION = 'core.anonymous_session'
    USER_AUTH = 'core.user_auth'
    TEMPORARY_USER_AUTH = 'core.user_auth.temporary'
    EMAIL_VERIFICATION = 'core.email_verification'
    PLUGINS_READ = 'core.plugins.read'
    SETTINGS_READ = 'core.setting.read'
    ASSETSTORES_READ = 'core.assetstore.read'
    PARTIAL_UPLOAD_READ = 'core.partial_upload.read'
    PARTIAL_UPLOAD_CLEAN = 'core.partial_upload.clean'
    DATA_READ = 'core.data.read'
    DATA_WRITE = 'core.data.write'
    DATA_OWN = 'core.data.own'
    USER_INFO_READ = 'core.user_info.read'

    _customScopes = []
    _adminCustomScopes = []
    _scopeIds = set()
    _adminScopeIds = set()

    @classmethod
    def describeScope(cls, scopeId, name, description, admin=False):
        """
        Register a description of a scope.

        :param scopeId: The unique identifier string for the scope.
        :type scopeId: str
        :param name: A short human readable name for the scope.
        :type name: str
        :param description: A more complete description of the scope.
        :type description: str
        :param admin: If this scope only applies to admin users, set to True.
        :type admin: bool
        """
        info = {
            'id': scopeId,
            'name': name,
            'description': description
        }
        if admin:
            cls._adminCustomScopes.append(info)
            cls._adminScopeIds.add(scopeId)
        else:
            cls._customScopes.append(info)
            cls._scopeIds.add(scopeId)

    @classmethod
    def listScopes(cls):
        return {
            'custom': cls._customScopes,
            'adminCustom': cls._adminCustomScopes
        }

    @classmethod
    def scopeIds(cls, admin=False):
        if admin:
            return cls._scopeIds | cls._adminScopeIds
        else:
            return cls._scopeIds


TokenScope.describeScope(
    TokenScope.USER_INFO_READ, 'Read your user information',
    'Allows clients to look up your user information, including private fields '
    'such as email address.')
TokenScope.describeScope(
    TokenScope.DATA_READ, 'Read data',
    'Allows clients to read all data that you have access to.')
TokenScope.describeScope(
    TokenScope.DATA_WRITE, 'Write data',
    'Allows clients to edit data in the hierarchy and create new data anywhere '
    'you have write access.')
TokenScope.describeScope(
    TokenScope.DATA_OWN, 'Data ownership', 'Allows administrative control '
    'on data you own, including setting access control and deletion.'
)

TokenScope.describeScope(
    TokenScope.PLUGINS_READ, 'See installed plugins', 'Allows clients '
    'to see the list of plugins installed on the server.', admin=True)
TokenScope.describeScope(
    TokenScope.SETTINGS_READ, 'See system setting values', 'Allows clients to '
    'view the value of any system setting.', admin=True)
TokenScope.describeScope(
    TokenScope.ASSETSTORES_READ, 'View assetstores', 'Allows clients to see '
    'all assetstore information.', admin=True)
TokenScope.describeScope(
    TokenScope.PARTIAL_UPLOAD_READ, 'View unfinished uploads.',
    'Allows clients to see all partial uploads.', admin=True)
TokenScope.describeScope(
    TokenScope.PARTIAL_UPLOAD_CLEAN, 'Remove unfinished uploads.',
    'Allows clients to remove unfinished uploads.', admin=True)


class CoreEventHandler(object):
    """
    This enum represents handler identifier strings for core event handlers.
    If you wish to unbind a core event handler, use one of these as the
    ``handlerName`` argument. Unbinding core event handlers can be used to
    disable certain default functionalities.
    """

    # For removing deleted user/group references from AccessControlledModel
    ACCESS_CONTROL_CLEANUP = 'core.cleanupDeletedEntity'

    # For updating an item's size to include a new file.
    FILE_PROPAGATE_SIZE = 'core.propagateSizeToItem'

    # For adding a group's creator into its ACL at creation time.
    GROUP_CREATOR_ACCESS = 'core.grantCreatorAccess'

    # For creating the default Public and Private folders at user creation time.
    USER_DEFAULT_FOLDERS = 'core.addDefaultFolders'

    # For adding a user into its own ACL.
    USER_SELF_ACCESS = 'core.grantSelfAccess'

    # For updating the cached webroot HTML when settings change.
    WEBROOT_SETTING_CHANGE = 'core.updateWebrootSettings'



# -*- coding: utf-8 -*-
import cherrypy
import datetime
import dateutil.parser
import errno
import json
import os
import pytz
import re
import string
import six

import girder
import girder.events

try:
    from random import SystemRandom
    random = SystemRandom()
    random.random()  # potentially raises NotImplementedError
except NotImplementedError:
    girder.logprint.warning(
        'WARNING: using non-cryptographically secure PRNG.')
    import random


def parseTimestamp(x, naive=True):
    """
    Parse a datetime string using the python-dateutil package.

    If no timezone information is included, assume UTC. If timezone information
    is included, convert to UTC.

    If naive is True (the default), drop the timezone information such that a
    naive datetime is returned.
    """
    dt = dateutil.parser.parse(x)
    if dt.tzinfo:
        dt = dt.astimezone(pytz.utc).replace(tzinfo=None)
    if naive:
        return dt
    else:
        return pytz.utc.localize(dt)


def genToken(length=64):
    """
    Use this utility function to generate a random string of a desired length.
    """
    return ''.join(random.choice(string.ascii_letters + string.digits)
                   for _ in range(length))


def camelcase(value):
    """
    Convert a module name or string with underscores and periods to camel case.

    :param value: the string to convert
    :type value: str
    :returns: the value converted to camel case.
    """
    return ''.join(x.capitalize() if x else '_' for x in
                   re.split('[._]+', value))


def mkdir(path, mode=0o777, recurse=True, existOk=True):
    """
    Create a new directory or ensure a directory already exists.

    :param path: The directory to create.
    :type path: str
    :param mode: The mode (permission mask) prior to applying umask.
    :type mode: int
    :param recurse: Whether intermediate missing dirs should be created.
    :type recurse: bool
    :param existOk: Set to True to suppress the error if the dir exists.
    :type existOk: bool
    """
    method = os.makedirs if recurse else os.mkdir

    try:
        method(path, mode)
    except OSError as exc:
        if existOk and exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def toBool(val):
    """
    Coerce a string value to a bool. Meant to be used to parse HTTP
    parameters, which are always sent as strings. The following string
    values will be interpreted as True:

      - ``'true'``
      - ``'on'``
      - ``'1'``
      - ``'yes'``

    All other strings will be interpreted as False. If the given param
    is not passed at all, returns the value specified by the default arg.
    This function is case-insensitive.

    :param val: The value to coerce to a bool.
    :type val: str
    """
    if isinstance(val, bool):
        return val

    return val.lower().strip() in ('true', 'on', '1', 'yes')


class JsonEncoder(json.JSONEncoder):
    """
    This extends the standard json.JSONEncoder to allow for more types to be
    sensibly serialized. This is used in Girder's REST layer to serialize
    route return values when JSON is requested.
    """

    def default(self, obj):
        event = girder.events.trigger('rest.json_encode', obj)
        if len(event.responses):
            return event.responses[-1]

        if isinstance(obj, set):
            return tuple(obj)
        elif isinstance(obj, datetime.datetime):
            return obj.replace(tzinfo=pytz.UTC).isoformat()
        return str(obj)


class RequestBodyStream(object):
    """
    Wraps a cherrypy request body into a more abstract file-like object.
    """

    _ITER_CHUNK_LEN = 65536

    def __init__(self, stream, size=None):
        self.stream = stream
        self.size = size

    def read(self, *args, **kwargs):
        return self.stream.read(*args, **kwargs)

    def close(self, *args, **kwargs):
        pass

    def __iter__(self):
        return self

    def __len__(self):
        return self.getSize()

    def __next__(self):
        data = self.read(self._ITER_CHUNK_LEN)
        if not data:
            raise StopIteration
        return data

    def next(self):
        return self.__next__()

    def getSize(self):
        """
        Returns the size of the body data wrapped by this class. For
        multipart encoding, this is the size of the part. For sending
        as the body, this is the Content-Length header.
        """
        if self.size is not None:
            return self.size

        return int(cherrypy.request.headers['Content-Length'])


def optionalArgumentDecorator(baseDecorator):
    """
    This decorator can be applied to other decorators, allowing the target decorator to be used
    either with or without arguments.

    The target decorator is expected to take at least 1 argument: the function to be wrapped. If
    additional arguments are provided by the final implementer of the target decorator, they will
    be passed to the target decorator as additional arguments.

    For example, this may be used as:

    .. code-block:: python

        @optionalArgumentDecorator
        def myDec(fun, someArg=None):
            ...

        @myDec
        def a(...):
            ...

        @myDec()
        def a(...):
            ...

        @myDec(5)
        def a(...):
            ...

        @myDec(someArg=5)
        def a(...):
            ...

    :param baseDecorator: The target decorator.
    :type baseDecorator: callable
    """
    @six.wraps(baseDecorator)
    def normalizedArgumentDecorator(*args, **kwargs):
        if len(args) == 1 and callable(args[0]):  # Applied as a raw decorator
            decoratedFunction = args[0]
            # baseDecorator must wrap and return decoratedFunction
            return baseDecorator(decoratedFunction)
        else:   # Applied as a argument-containing decorator
            # Decoration will occur in two passes:
            #   * Now, the decorator arguments are passed, and a new decorator should be returned
            #   * Afterwards, the new decorator will be called to decorate the decorated function
            decoratorArgs = args
            decoratorKwargs = kwargs

            def partiallyAppliedDecorator(decoratedFunction):
                return baseDecorator(decoratedFunction, *decoratorArgs, **decoratorKwargs)
            return partiallyAppliedDecorator

    return normalizedArgumentDecorator



import collections
import os
import six
import subprocess
import re

Tree = lambda: collections.defaultdict(Tree)  # noqa: E731

EXCLUDE_DIRS = [
    # Only look in girder, plugins and clients folders i.e. exclude all directories that don't begin
    # with "clients" or "girder" or "plugins"
    '^(?!(clients|girder|plugins|pytest_girder))',
    # Exclude plugin tests
    'plugin_tests',
    '\\.egg/',
    'node_modules/']

IGNORE_FILES = ['setup.py']

excluder = re.compile('|'.join(EXCLUDE_DIRS))

baseTree = Tree()


def addSymbol(symbolScope, symbolTree):
    if not symbolScope:
        return
    addSymbol(
        symbolScope[1:],
        symbolTree[symbolScope[0]]
    )


def addFileSymbols(filePath, symbolTree):
    fileTags = subprocess.check_output([
        'ctags',
        '-f', '-',
        '--languages=python',
        '--python-kinds=%s' % ''.join([
            # Skip imported symbols
            '-i',
            # Skip "as"-renamed imported modules
            '-I'
            # Skip unknown symbols (which are typically "as"-renamed imported symbols)
            '-x',

        ]),
        filePath
    ]).decode('utf-8')

    for fileTag in fileTags.splitlines():
        symbolName, symbolFileName, symbolRegex, symbolExtensions = fileTag.split('\t', 3)
        symbolExtensions = symbolExtensions.split('\t')
        symbolAbbrviatedKind = symbolExtensions[0]  # noqa: F841
        if len(symbolExtensions) >= 2:
            # Symbol is scoped
            symbolScopeKind, symbolScope = symbolExtensions[1].split(':')
            if symbolScopeKind == 'class':
                pass
            elif symbolScopeKind == 'function':
                # Symbols defined inside functions are not visible
                continue
            elif symbolScopeKind == 'member':
                # Symbols defined inside methods are not visible
                continue
            else:
                raise Exception('Unknown symbol scope "%s" in %s' % (symbolScopeKind, filePath))
            symbolScope = symbolScope.split('.')
        else:
            symbolScope = []
        symbolScope.append(symbolName)

        addSymbol(symbolScope, symbolTree)


def addDirSymbols(dirPath, symbolTree):
    for subName in os.listdir(dirPath):
        subPath = os.path.join(dirPath, subName)
        if os.path.isfile(subPath):
            subNameBase, subNameExt = os.path.splitext(subName)
            if subNameExt != '.py' or subName in IGNORE_FILES:
                continue
            if subName == '__init__.py':
                # '__init__.py' adds symbols to the module-level tree
                addFileSymbols(subPath, symbolTree)
            else:
                addFileSymbols(subPath, symbolTree[subNameBase])


def printTree(symbolTree, level=0):
    for symbol, subTree in sorted(six.viewitems(symbolTree)):
        if symbol.startswith('_') and symbol != '__init__.py':
            continue
        print(' ' * (level * 4) + symbol)
        printTree(subTree, level + 1)


def _get_subtree_hierarchy(tree, symbol_list):
    if symbol_list == []:
        return tree

    return _get_subtree_hierarchy(tree[symbol_list.pop(0)], symbol_list)


def _root_relative_path(directory_path, rootPath):
    return directory_path[len(rootPath) + 1:]


def _valid_directory(dp, f, rootPath):
    return os.path.splitext(f)[1] == '.py' and \
        _root_relative_path(dp, rootPath) != '' and \
        not excluder.search(_root_relative_path(dp, rootPath))


def main():
    rootPath = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))

    # Note: that's a set comprehension not a dict comprehension
    module_or_package_directories = {_root_relative_path(dp, rootPath)
                                     for dp, _, fn in os.walk(rootPath) for f in fn
                                     if _valid_directory(dp, f, rootPath)}

    for d in module_or_package_directories:
        addDirSymbols(
            os.path.join(rootPath, d),
            _get_subtree_hierarchy(baseTree, d.split(os.sep)))

    printTree(baseTree)


if __name__ == '__main__':
    main()



"""
This script is for deleting old audit log entries from the database. Example invocation:

    girder audit-logs-cleanup --days=30 --types=rest.request
"""

import click
import datetime
from girder_audit_logs import Record


@click.command(name='audit-logs-cleanup')
@click.option('--days', type=click.INT, help='How many days to preserve records', default=90)
@click.option('--types', help='Which record types to remove as a comma separated list. If not '
              'provided, removes all record types.')
def cleanup(days, types):
    filter = {'when': {'$lt': datetime.datetime.utcnow() - datetime.timedelta(days=days)}}

    if types:
        filter['type'] = {'$in': types.split(',')}

    click.echo('Deleted %d log entries.' % Record().collection.delete_many(filter).deleted_count)


if __name__ == '__main__':
    cleanup()



# -*- coding: utf-8 -*-
import cherrypy
import datetime
import errno
import girder
import json
import six
import os
import logging

from girder import plugin
from girder.api import access
from girder.constants import TokenScope, ACCESS_FLAGS, VERSION
from girder.exceptions import GirderException, ResourcePathNotFound
from girder.models.collection import Collection
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.group import Group
from girder.models.item import Item
from girder.models.setting import Setting
from girder.models.upload import Upload
from girder.models.user import User
from girder.settings import SettingKey
from girder.utility import config, system
from girder.utility.progress import ProgressContext
from ..describe import Description, autoDescribeRoute
from ..rest import Resource

ModuleStartTime = datetime.datetime.utcnow()
LOG_BUF_SIZE = 65536


class System(Resource):
    """
    The system endpoints are for querying and managing system-wide properties.
    """

    def __init__(self):
        super(System, self).__init__()
        self.resourceName = 'system'
        self.route('DELETE', ('setting',), self.unsetSetting)
        self.route('GET', ('version',), self.getVersion)
        self.route('GET', ('configuration',), self.getConfigurationOption)
        self.route('GET', ('setting',), self.getSetting)
        self.route('GET', ('plugins',), self.getPlugins)
        self.route('GET', ('access_flag',), self.getAccessFlags)
        self.route('PUT', ('setting',), self.setSetting)
        self.route('GET', ('uploads',), self.getPartialUploads)
        self.route('DELETE', ('uploads',), self.discardPartialUploads)
        self.route('GET', ('check',), self.systemStatus)
        self.route('PUT', ('check',), self.systemConsistencyCheck)
        self.route('GET', ('log',), self.getLog)
        self.route('GET', ('log', 'level'), self.getLogLevel)
        self.route('PUT', ('log', 'level'), self.setLogLevel)
        self.route('GET', ('setting', 'collection_creation_policy', 'access'),
                   self.getCollectionCreationPolicyAccess)

    @access.admin
    @autoDescribeRoute(
        Description('Set the value for a system setting, or a list of them.')
        .notes('Must be a system administrator to call this. If the value passed is '
               'a valid JSON object, it will be parsed and stored as an object.')
        .param('key', 'The key identifying this setting.', required=False, paramType='formData')
        .param('value', 'The value for this setting.', required=False, paramType='formData')
        .jsonParam('list', 'A JSON list of objects with key and value representing '
                   'a list of settings to set.', required=False, requireArray=True,
                   paramType='formData')
        .errorResponse('You are not a system administrator.', 403)
        .errorResponse('Failed to set system setting.', 500)
    )
    def setSetting(self, key, value, list):
        if list is None:
            list = ({'key': key, 'value': value},)

        for setting in list:
            key, value = setting['key'], setting['value']
            if isinstance(value, six.string_types):
                try:
                    value = json.loads(value)
                except ValueError:
                    pass

            if value is None:
                Setting().unset(key=key)
            else:
                Setting().set(key=key, value=value)

        return True

    @access.admin(scope=TokenScope.SETTINGS_READ)
    @autoDescribeRoute(
        Description('Get the value of a system configuration option.')
        .notes('Must be a system administrator to call this.')
        .param('section', 'The section identifying the configuration option.', required=True)
        .param('key', 'The key identifying the configuration option.', required=True)
        .errorResponse('You are not a system administrator.', 403)
        .errorResponse('No such option with the given section/key exists.', 404)
    )
    def getConfigurationOption(self, section, key, params):
        configSection = config.getConfig().get(section)

        if configSection is None:
            raise ResourcePathNotFound('No section with that name exists.')
        elif key not in configSection:
            raise ResourcePathNotFound('No key with that name exists.')
        else:
            return configSection.get(key)

    @access.admin(scope=TokenScope.SETTINGS_READ)
    @autoDescribeRoute(
        Description('Get the value of a system setting, or a list of them.')
        .notes('Must be a system administrator to call this.')
        .param('key', 'The key identifying this setting.', required=False)
        .jsonParam('list', 'A JSON list of keys representing a set of settings to return.',
                   required=False, requireArray=True)
        .errorResponse('You are not a system administrator.', 403)
    )
    def getSetting(self, key, list):
        if list is not None:
            return {k: Setting().get(k) for k in list}
        else:
            self.requireParams({'key': key})
            return Setting().get(key)

    @access.admin(scope=TokenScope.PLUGINS_READ)
    @autoDescribeRoute(
        Description('Get the lists of all available and all loaded plugins.')
        .notes('Must be a system administrator to call this.')
        .errorResponse('You are not a system administrator.', 403)
    )
    def getPlugins(self):
        def _pluginNameToResponse(name):
            p = plugin.getPlugin(name)
            return {
                'name': p.displayName,
                'description': p.description,
                'url': p.url,
                'version': p.version
            }

        return {
            'all': {name: _pluginNameToResponse(name) for name in plugin.allPlugins()},
            'loaded': plugin.loadedPlugins()
        }

    @access.public
    @autoDescribeRoute(
        Description('Get the version information for this server.')
    )
    def getVersion(self):
        version = dict(**VERSION)
        version['serverStartDate'] = ModuleStartTime
        return version

    @access.admin
    @autoDescribeRoute(
        Description('Unset the value for a system setting.')
        .notes('Must be a system administrator to call this. This is used to '
               'explicitly restore a setting to its default value.')
        .param('key', 'The key identifying the setting to unset.')
        .errorResponse('You are not a system administrator.', 403)
    )
    def unsetSetting(self, key):
        return Setting().unset(key)

    @access.admin(scope=TokenScope.PARTIAL_UPLOAD_READ)
    @autoDescribeRoute(
        Description('Get a list of uploads that have not been finished.')
        .notes('Must be a system administrator to call this.')
        .param('uploadId', 'List only a specific upload.', required=False)
        .param('userId', 'Restrict listing uploads to those started by a '
               'specific user.', required=False)
        .param('parentId', 'Restrict listing uploads to those within a '
               'specific folder or item.', required=False)
        .param('assetstoreId', 'Restrict listing uploads within a specific assetstore.',
               required=False)
        .param('minimumAge', 'Restrict listing uploads to those that are at '
               'least this many days old.', dataType='float', required=False)
        .param('includeUntracked', 'Some assetstores can have partial uploads '
               'that are no longer in the Girder database.  If this is True, '
               'include all of them (only filtered by assetstoreId) in the '
               'result list.', required=False, dataType='boolean', default=True)
        .pagingParams(defaultSort='updated')
        .errorResponse('You are not a system administrator.', 403)
    )
    def getPartialUploads(self, uploadId, userId, parentId, assetstoreId, minimumAge,
                          includeUntracked, limit, offset, sort):
        filters = {}
        if uploadId is not None:
            filters['uploadId'] = uploadId
        if userId is not None:
            filters['userId'] = userId
        if assetstoreId is not None:
            filters['assetstoreId'] = assetstoreId
        if parentId is not None:
            filters['parentId'] = parentId
        if minimumAge is not None:
            filters['minimumAge'] = minimumAge

        uploadList = list(Upload().list(
            filters=filters, limit=limit, offset=offset, sort=sort))
        if includeUntracked and (limit == 0 or len(uploadList) < limit):
            untrackedList = Upload().untrackedUploads('list', assetstoreId)
            if limit == 0:
                uploadList += untrackedList
            elif len(uploadList) < limit:
                uploadList += untrackedList[:limit - len(uploadList)]
        return uploadList

    @access.admin(scope=TokenScope.PARTIAL_UPLOAD_CLEAN)
    @autoDescribeRoute(
        Description('Discard uploads that have not been finished.')
        .notes('Must be a system administrator to call this. This frees '
               'resources that were allocated for the uploads and clears the '
               'uploads from database.')
        .param('uploadId', 'Clear only a specific upload.', required=False)
        .param('userId', 'Restrict clearing uploads to those started by a '
               'specific user.', required=False)
        .param('parentId', 'Restrict clearing uploads to those within a '
               'specific folder or item.', required=False)
        .param('assetstoreId', 'Restrict clearing uploads within a specific assetstore.',
               required=False)
        .param('minimumAge', 'Restrict clearing uploads to those that are at '
               'least this many days old.', dataType='float', required=False)
        .param('includeUntracked', 'Some assetstores can have partial uploads '
               'that are no longer in the Girder database.  If this is True, '
               'remove all of them (only filtered by assetstoreId).',
               required=False, dataType='boolean', default=True)
        .errorResponse('You are not a system administrator.', 403)
        .errorResponse('Failed to delete upload', 500)
    )
    def discardPartialUploads(self, uploadId, userId, parentId, assetstoreId,
                              minimumAge, includeUntracked):
        filters = {}
        if uploadId is not None:
            filters['uploadId'] = uploadId
        if userId is not None:
            filters['userId'] = userId
        if assetstoreId is not None:
            filters['assetstoreId'] = assetstoreId
        if parentId is not None:
            filters['parentId'] = parentId
        if minimumAge is not None:
            filters['minimumAge'] = minimumAge
        uploadList = list(Upload().list(filters=filters))
        # Move the results to list that isn't a cursor so we don't have to have
        # the cursor sitting around while we work on the data.
        for upload in uploadList:
            try:
                Upload().cancelUpload(upload)
            except OSError as exc:
                if exc.errno == errno.EACCES:
                    raise GirderException(
                        'Failed to delete upload.',
                        'girder.api.v1.system.delete-upload-failed')
                raise
        if includeUntracked:
            uploadList += Upload().untrackedUploads('delete', assetstoreId)
        return uploadList

    @access.public
    @autoDescribeRoute(
        Description('Report the current system status.')
        .notes('Must be a system administrator to call this with any mode '
               'other than basic.')
        .param('mode', 'Select details to return. "quick" are the details '
               'that can be answered without much load on the system. "slow" '
               'also includes some resource-intensive queries.',
               required=False, enum=('basic', 'quick', 'slow'), default='basic')
        .errorResponse('You are not a system administrator.', 403)
    )
    def systemStatus(self, mode):
        user = self.getCurrentUser()
        if mode != 'basic':
            self.requireAdmin(user)
        status = system.getStatus(mode, user)
        status['requestBase'] = cherrypy.request.base.rstrip('/')
        return status

    @access.public
    @autoDescribeRoute(Description('List all access flags available in the system.'))
    def getAccessFlags(self):
        return ACCESS_FLAGS

    @access.admin
    @autoDescribeRoute(
        Description('Perform a variety of system checks to verify that all is '
                    'well.')
        .notes('Must be a system administrator to call this.  This verifies '
               'and corrects some issues, such as incorrect folder sizes.')
        .param('progress', 'Whether to record progress on this task.',
               required=False, dataType='boolean', default=False)
        .errorResponse('You are not a system administrator.', 403)
    )
    def systemConsistencyCheck(self, progress):
        user = self.getCurrentUser()
        title = 'Running system consistency check'
        with ProgressContext(progress, user=user, title=title) as pc:
            results = {}
            pc.update(title='Checking for orphaned records (Step 1 of 3)')
            results['orphansRemoved'] = self._pruneOrphans(pc)
            pc.update(title='Checking for incorrect base parents (Step 2 of 3)')
            results['baseParentsFixed'] = self._fixBaseParents(pc)
            pc.update(title='Checking for incorrect sizes (Step 3 of 3)')
            results['sizesChanged'] = self._recalculateSizes(pc)
            return results
        # TODO:
        # * check that all files are associated with an existing item
        # * check that all files exist within their assetstore and are the
        #   expected size
        # * check that all folders have a valid ancestor tree leading to a
        #   user or collection
        # * check that all folders have the correct baseParentId and
        #   baseParentType
        # * check that all groups contain valid users
        # * check that all resources validate
        # * for filesystem assetstores, find files that are not tracked.
        # * for gridfs assetstores, find chunks that are not tracked.
        # * for s3 assetstores, find elements that are not tracked.

    @access.admin
    @autoDescribeRoute(
        Description('Show the most recent contents of the server logs.')
        .notes('Must be a system administrator to call this.')
        .param('bytes', 'Controls how many bytes (from the end of the log) to show. '
               'Pass 0 to show the whole log.', dataType='integer', required=False, default=4096)
        .param('log', 'Which log to tail.', enum=('error', 'info'),
               required=False, default='error')
        .errorResponse('You are not a system administrator.', 403)
    )
    def getLog(self, bytes, log):
        path = girder.getLogPaths()[log]
        filesize = os.path.getsize(path)
        length = int(bytes) or filesize
        filesize1 = 0
        if length > filesize:
            path1 = path + '.1'
            if os.path.exists(path1):
                filesize1 = os.path.getsize(path1)

        def stream():
            yield '=== Last %d bytes of %s: ===\n\n' % (
                min(length, filesize + filesize1), path
            )

            readlength = length
            if readlength > filesize and filesize1:
                readlength = length - filesize
                with open(path1, 'rb') as f:
                    if readlength < filesize1:
                        f.seek(-readlength, os.SEEK_END)
                    while True:
                        data = f.read(LOG_BUF_SIZE)
                        if not data:
                            break
                        yield data
                readlength = filesize
            with open(path, 'rb') as f:
                if readlength < filesize:
                    f.seek(-readlength, os.SEEK_END)
                while True:
                    data = f.read(LOG_BUF_SIZE)
                    if not data:
                        break
                    yield data
        return stream

    @access.admin
    @autoDescribeRoute(
        Description('Get the current log level.')
        .notes('Must be a system administrator to call this.')
        .errorResponse('You are not a system administrator.', 403)
    )
    def getLogLevel(self):
        level = girder.logger.getEffectiveLevel()
        return logging.getLevelName(level)

    @access.admin
    @autoDescribeRoute(
        Description('Set the current log level.')
        .notes('Must be a system administrator to call this.')
        .param('level', 'The new level to set.',
               enum=['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'],
               default='INFO')
        .errorResponse('You are not a system administrator.', 403)
    )
    def setLogLevel(self, level):
        level = logging.getLevelName(level)
        for logger in (girder.logger, cherrypy.log.access_log, cherrypy.log.error_log):
            logger.setLevel(level)
            for handler in logger.handlers:
                handler.setLevel(level)
        return logging.getLevelName(level)

    @access.admin
    @autoDescribeRoute(
        Description('Get access of content creation policy.')
        .notes('Get result in the same structure as the access endpoints '
               'of collection, file, and group')
    )
    def getCollectionCreationPolicyAccess(self):
        cpp = Setting().get(SettingKey.COLLECTION_CREATE_POLICY)

        acList = {
            'users': [{'id': x} for x in cpp.get('users', [])],
            'groups': [{'id': x} for x in cpp.get('groups', [])]
        }

        for user in acList['users'][:]:
            userDoc = User().load(
                user['id'], force=True,
                fields=['firstName', 'lastName', 'login'])
            if userDoc is None:
                acList['users'].remove(user)
            else:
                user['login'] = userDoc['login']
                user['name'] = ' '.join((userDoc['firstName'], userDoc['lastName']))

        for grp in acList['groups'][:]:
            grpDoc = Group().load(
                grp['id'], force=True, fields=['name', 'description'])
            if grpDoc is None:
                acList['groups'].remove(grp)
            else:
                grp['name'] = grpDoc['name']
                grp['description'] = grpDoc['description']

        return acList

    def _fixBaseParents(self, progress):
        fixes = 0
        models = [Folder(), Item()]
        steps = sum(model.find().count() for model in models)
        progress.update(total=steps, current=0)
        for model in models:
            for doc in model.find():
                progress.update(increment=1)
                baseParent = model.parentsToRoot(doc, force=True)[0]
                baseParentType = baseParent['type']
                baseParentId = baseParent['object']['_id']
                if (doc['baseParentType'] != baseParentType or doc['baseParentId'] != baseParentId):
                    model.update({'_id': doc['_id']}, update={
                        '$set': {
                            'baseParentType': baseParentType,
                            'baseParentId': baseParentId
                        }})
                    fixes += 1
        return fixes

    def _pruneOrphans(self, progress):
        count = 0
        models = [File(), Folder(), Item()]
        steps = sum(model.find().count() for model in models)
        progress.update(total=steps, current=0)
        for model in models:
            for doc in model.find():
                progress.update(increment=1)
                if model.isOrphan(doc):
                    model.remove(doc)
                    count += 1
        return count

    def _recalculateSizes(self, progress):
        fixes = 0
        models = [Collection(), User()]
        steps = sum(model.find().count() for model in models)
        progress.update(total=steps, current=0)
        for model in models:
            for doc in model.find():
                progress.update(increment=1)
                _, f = model.updateSize(doc)
                fixes += f
        return fixes
